// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <gtest/gtest.h>
#include "kde/gaussian.h"
#include "kde/explanation.h"

namespace xokdepp {

  typedef gaussian<vector_type> gaussian_type;
  typedef explanation<gaussian_type, vector_type> explanation_type;

  class explanation_gaussian_diagonal_test: public testing::Test {
  protected:
    virtual void SetUp() {
      //initialization will depend on the test
    }

  };

  static constexpr size_t dims = 3;

  TEST_F(explanation_gaussian_diagonal_test,EmptyConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Zero(dims);
    const explanation_type g1(dims);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(explanation_gaussian_diagonal_test,MeanConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Zero(dims);
    const explanation_type g1(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(explanation_gaussian_diagonal_test,MeanCovarianceConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Zero(dims);
    const explanation_type g1(v1, c1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance())<< "Ref: " << std::endl << c1 << std::endl << "Actual:" << std::endl << g1.covariance();
  }

  TEST_F(explanation_gaussian_diagonal_test,MeanCovarianceConstructor2) {
    //in this test covariance is diagonal, such that cov is invertible
    const vector_type v1 = vector_type::Constant(dims, 1);
    const vector_type c_zero = vector_type::Zero(dims);
    const vector_type c_diag_1 = v1;
    const explanation_type g1(v1, c_diag_1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c_diag_1, g1.covariance())<< "Ref: " << std::endl << c_zero << std::endl << "Actual:" << std::endl << g1.covariance();
  }

  TEST_F(explanation_gaussian_diagonal_test,MeanCovarianceConstructor3) {
    //in this test covariance is non-invertible (but still positive semi-definite)
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Constant(dims, 1);
    vector_type c2(dims);
    c2 << 1.000000000000, 1.000000000000, 1.000000000000;
    const explanation_type g1(v1, c1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_TRUE(allclose(c2, g1.covariance(), xokdepp::epsilon12, xokdepp::epsilon))<< "Ref: " << std::endl << c2 << std::endl << "Actual:" << std::endl << g1.covariance();
  }

  TEST_F(explanation_gaussian_diagonal_test,MeanCovarianceConstructor4) {
      //in this test covariance is non-invertible (but still positive semi-definite)
      const vector_type v1 = vector_type::Zero(dims);
      vector_type c1(dims);
      c1 << 1, 0, 1;
      vector_type c2(dims);
      c2 << 1.000000000000, 0.010000000000, 1.000000000000;
      const explanation_type g1(v1, c1);
      ASSERT_EQ(v1, g1.mean());
      ASSERT_TRUE(allclose(c2, g1.covariance(), xokdepp::epsilon12, xokdepp::epsilon))<< "Ref: " << std::endl << c2 << std::endl << "Actual:" << std::endl << g1.covariance();
    }

  TEST_F(explanation_gaussian_diagonal_test,merge_explanation1) {
    //this test merges two explanations into one. Checks if mean and covariance is correct, but also the underlying model
    const vector_type v0 = vector_type::Zero(dims);
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;

    const vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims);
    c_diag_2 << 1.3, 3.4, 23;

    //make reference gaussian - this should be the result of moment match of all explanations
    mixture<gaussian_type> ref_m(dims);
    ref_m.add( { gaussian_type(v0, c_diag_1), gaussian_type(v1, c_diag_1) });
    ref_m.normalize_weights();

    gaussian_type ref_g(ref_m, c_diag_1);

    mixture<explanation_type> m1(dims);
    m1.add( { explanation_type(v0, c_diag_1), explanation_type(v1, c_diag_1) });
    m1.normalize_weights();

    explanation_type e2(m1, c_diag_1);

    ASSERT_EQ(ref_g.mean(), e2.mean());
    ASSERT_EQ(v0, e2.detailed_model().component(0).mean());
    ASSERT_EQ(v1, e2.detailed_model().component(1).mean());
  }

  TEST_F(explanation_gaussian_diagonal_test,degenerate_covariance_correction) {
    //in this test covariance is diagonal, such that cov is invertible
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type c1(dims), c_ref(dims);
    c1 << 1.3, 0, 23;
    c_ref << 1.3000000001, 0.121500000001, 23.0000000001;

    explanation_type g1(v1, c1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_TRUE(allclose(c_ref, g1.covariance(), xokdepp::epsilon12, xokdepp::epsilon))<<"Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl << g1.covariance();
}

}
