// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <gtest/gtest.h>
#include "kde/gaussian.h"

namespace xokdepp {

  typedef gaussian<vector_type> gaussian_type;

  class gaussian_diagonal_test: public testing::Test {

  protected:
    virtual void SetUp() {
      //initialization will depend on the test
    }

  };

  static constexpr size_t dims = 3;

  TEST_F(gaussian_diagonal_test,EmptyConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Zero(dims);
    const gaussian_type g1(dims);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(gaussian_diagonal_test,EmptyConstructor2) {
    //test attributions after empty constructor
    const vector_type v1 = vector_type::Constant(dims, 1);
    const vector_type c1 = vector_type::Zero(dims);
    gaussian_type g1(dims);
    g1.mean(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(gaussian_diagonal_test,MeanConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Zero(dims);
    const gaussian_type g1(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(gaussian_diagonal_test,MeanCovarianceConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const vector_type c1 = vector_type::Zero(dims);
    const gaussian_type g1(v1, c1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance())<< "Ref: " << std::endl << c1 << std::endl << "Actual:" << std::endl << g1.covariance();
  }

  TEST_F(gaussian_diagonal_test,MeanCovarianceConstructor2) {
    //in this test covariance is diagonal, such that cov is invertible
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type c_diag_1 = v1;
    const gaussian_type g1(v1, c_diag_1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c_diag_1, g1.covariance())<< "Ref: " << std::endl << c_diag_1 << std::endl << "Actual:" << std::endl << g1.covariance();
    ASSERT_EQ(c_diag_1, g1.covariance());
  }

  TEST_F(gaussian_diagonal_test,MeanCovarianceConstructor3) {
    //in this test covariance is diagonal, such that cov is invertible
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type c_ref(dims);
    c_ref << 1.25, 1.25, 1.25;
    vector_type c_ref_diag(dims, dims);
    c_ref_diag << 1.25, 1.25, 1.25;
    const gaussian_type g1(v1, c_ref);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c_ref, g1.covariance())<< "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl << g1.covariance();
    ASSERT_EQ(c_ref_diag, g1.covariance());
  }

  TEST_F(gaussian_diagonal_test,moment_match_constructor) {
    const vector_type v0 = vector_type::Zero(dims);
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v_ref(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v_ref << 0.5, 0.5, 0.5;

    vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims);
    c_diag_2 << 1.3, 3.4, 23;

    vector_type c_ref(dims);
    c_ref << 1.25, 1.25, 1.25;

    vector_type v_ref2(dims), v_ref3(dims);
    v_ref2 << 6.6, 6, 23.5;
    v_ref3 << 0.733333333333333, 4.33333333333333, 48;
    vector_type c_ref2(dims), c_ref3(dims);
    c_ref2 << 42.11, 2.2, 564.25;
    c_ref3 << 1.34222222222222, 8.15555555555556, 4528.33333333333;

    //first case
    gaussian_type g1(v0, c_diag_1), g2(v1, c_diag_1);
    mixture<gaussian_type> m1(dims);
    m1.add(g1);
    m1.add(g2);
    m1.normalize_weights();

    gaussian_type g_m(m1, c_diag_2);

    ASSERT_EQ(v_ref, g_m.mean())<< "Ref: " << std::endl << v_ref << std::endl << "Actual:" << std::endl << g_m.mean();
    ASSERT_EQ(c_ref, g_m.covariance())<< "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl << g_m.covariance();

    //second case
    gaussian_type g3(v3, c_diag_1), g4(v4, c_diag_2);
    mixture<gaussian_type> m2(dims);
    m2.add(g3);
    m2.add(g4);
    m2.normalize_weights();

    gaussian_type g_m2(m2, c_diag_1);

    vector_type cd_ref2 = vector_type(dims);
    cd_ref2 << 42.11, 2.2, 564.25;
    ASSERT_TRUE(allclose(v_ref2, g_m2.mean(), xokdepp::epsilon12, xokdepp::epsilon));
    ASSERT_TRUE(allclose(cd_ref2, g_m2.covariance().diagonal(),xokdepp::epsilon12,xokdepp::epsilon))<< "Ref: " << std::endl << cd_ref2 << std::endl << "Actual:" << std::endl << g_m2.covariance().diagonal();

    //third case
    gaussian_type g5(v4, c_diag_1), g6(v2, c_diag_2), g7(v1, c_diag_2);
    mixture<gaussian_type> m3(dims);
    m3.add(g5);
    m3.add(g6);
    m3.add(g7);
    m3.normalize_weights();

    gaussian_type g_m3(m3, c_diag_1);

    vector_type cd_ref3 = vector_type(dims);
    cd_ref3 << 1.34222222222222, 8.15555555555556, 4528.33333333333;
    ASSERT_TRUE(allclose(v_ref3, g_m3.mean(), xokdepp::epsilon12, xokdepp::epsilon));
    ASSERT_TRUE(allclose(cd_ref3, g_m3.covariance().diagonal(), xokdepp::epsilon12, xokdepp::epsilon))<< "Ref: " << std::endl << cd_ref3 << std::endl << "Actual:" << std::endl << g_m3.covariance().diagonal();;
  }

  TEST_F(gaussian_diagonal_test,Reset) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type c0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type c1 = vector_type::Constant(dims, 1);
    gaussian_type g1(v1, c1);
    g1.reset_mean();
    g1.reset_covariance();
    ASSERT_EQ(v0, g1.mean());
    vector_type cd0 = vector_type::Zero(dims);
    ASSERT_EQ(cd0, g1.covariance())<< "Ref: " << std::endl << cd0 << std::endl << "Actual:" << std::endl << g1.covariance();

  }

  TEST_F(gaussian_diagonal_test,Probability) {
    //<< "Ref: " << std::endl << -std::numeric_limits<data_type>::infinity() << std::endl << "Actual:" << std::endl << g1.log_likelihood(obs2);
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type c_diag_1 = v1;
    gaussian_type g1(v0, c_diag_1);

    vector_type obs1(dims), obs2(dims), obs3(dims), obs4(dims);
    obs1 << 1, 1, 1;
    ASSERT_NEAR(0.014167345154413, g1.likelihood(obs1), xokdepp::epsilon);
    ASSERT_NEAR(-4.256815599614018, g1.log_likelihood(obs1), xokdepp::epsilon);
    obs2 << 18, 21, 293;
    ASSERT_EQ(0, g1.likelihood(obs2));
    ASSERT_NEAR(-43309.75681559961, g1.log_likelihood(obs2), xokdepp::epsilon);
    obs3 << 1, 0, 0;
    ASSERT_NEAR(0.038510836890749, g1.likelihood(obs3), xokdepp::epsilon);
    ASSERT_NEAR(-3.256815599614018, g1.log_likelihood(obs3), xokdepp::epsilon);
    obs4 << 0, 0, 0;
    ASSERT_NEAR(0.063493635934241, g1.likelihood(obs4), xokdepp::epsilon);
    ASSERT_NEAR(-2.756815599614018, g1.log_likelihood(obs4), xokdepp::epsilon);
  }

  TEST_F(gaussian_diagonal_test,Copy) {
    const vector_type v1 = vector_type::Constant(dims, 1);
    const vector_type c_diag_1 = v1;
    const gaussian_type g1(v1, c_diag_1);
    const gaussian_type g2 = g1;

    ASSERT_EQ(g1.mean(), g2.mean());
    ASSERT_EQ(v1, g2.mean());
    ASSERT_EQ(g1.covariance(), g2.covariance());
    ASSERT_EQ(c_diag_1, g1.covariance());

  }

  TEST_F(gaussian_diagonal_test,modify_gaussian_diagonal) {
    //this test initializes a gaussian_type, and then changes its mean and covariance.
    //the data used to initialize the gaussian_type should be untouched
    vector_type v_ref = vector_type::Zero(dims);
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type c_diag_1 = v1;
    gaussian_type g1(v0, c_diag_1);

    ASSERT_EQ(v0, g1.mean());
    g1.mean(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(v_ref, v0);
  }

  TEST_F(gaussian_diagonal_test,KL_divergence) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;

    vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims);
    c_diag_2 << 1.3, 3.4, 23;

    gaussian_type g0(v0, c_diag_1);
    gaussian_type g1(v1, c_diag_1);
    gaussian_type g2(v2, c_diag_1);
    gaussian_type g3(v3, c_diag_1);
    gaussian_type g4(v4, c_diag_1);

    ASSERT_EQ(0, g0.KL_divergence(g0));
    ASSERT_EQ(1.5, g0.KL_divergence(g1));
    ASSERT_EQ(10243, g0.KL_divergence(g2));
    ASSERT_EQ(1207, g0.KL_divergence(g3));
    ASSERT_EQ(18.02, g0.KL_divergence(g4));

    ASSERT_EQ(1.5, g1.KL_divergence(g0));
    ASSERT_EQ(0, g1.KL_divergence(g1));
    ASSERT_EQ(10094.5, g1.KL_divergence(g2));
    ASSERT_EQ(1142.5, g1.KL_divergence(g3));
    ASSERT_EQ(13.32, g1.KL_divergence(g4));

    ASSERT_EQ(10243, g2.KL_divergence(g0));
    ASSERT_EQ(10094.5, g2.KL_divergence(g1));
    ASSERT_EQ(0, g2.KL_divergence(g2));
    ASSERT_EQ(4680, g2.KL_divergence(g3));
    ASSERT_NEAR(10224.819999999999709, g2.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_EQ(1207, g3.KL_divergence(g0));
    ASSERT_EQ(1142.5, g3.KL_divergence(g1));
    ASSERT_EQ(4680, g3.KL_divergence(g2));
    ASSERT_EQ(0, g3.KL_divergence(g3));
    ASSERT_NEAR(1186.420000000000073, g3.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_EQ(18.02, g4.KL_divergence(g0));
    ASSERT_EQ(13.32, g4.KL_divergence(g1));
    ASSERT_NEAR(10224.819999999999709, g4.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(1186.420000000000073, g4.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_EQ(0, g4.KL_divergence(g4));

    //gaussian_diagonals with non diag-only covariance
    gaussian_type g_0(v0, c_diag_2);
    gaussian_type g_1(v1, c_diag_2);
    gaussian_type g_2(v2, c_diag_2);
    gaussian_type g_3(v3, c_diag_2);
    gaussian_type g_4(v4, c_diag_2);

    ASSERT_EQ(0, g_0.KL_divergence(g_0));
    EXPECT_NEAR(0.55341333857957897, g_0.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(450.22221129254376, g_0.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(118.31585677749359, g_0.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(5.3095022624434387, g_0.KL_divergence(g_4), xokdepp::epsilon);

    EXPECT_NEAR(0.55341333857957897, g_1.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_EQ(0, g_1.KL_divergence(g_1));
    EXPECT_NEAR(442.02429667519181, g_1.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(105.06108597285068, g_1.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(3.9443635648239233, g_1.KL_divergence(g_4), xokdepp::epsilon);

    EXPECT_NEAR(450.22221129254376, g_2.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(442.02429667519181, g_2.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_EQ(0, g_2.KL_divergence(g_2));
    EXPECT_NEAR(255.73244147157189, g_2.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(444.78963210702341, g_2.KL_divergence(g_4), xokdepp::epsilon);

    EXPECT_NEAR(118.31585677749359, g_3.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(105.06108597285068, g_3.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(255.73244147157189, g_3.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_EQ(0, g_3.KL_divergence(g_3));
    EXPECT_NEAR(111.0371237458194, g_3.KL_divergence(g_4), xokdepp::epsilon);

    EXPECT_NEAR(5.3095022624434387, g_4.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(3.9443635648239233, g_4.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(444.78963210702341, g_4.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(111.0371237458194, g_4.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_EQ(0, g_4.KL_divergence(g_4));

    //now mixed comparison
    EXPECT_NEAR(1.3642302945889573, g0.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(10.039183043990622, g_0.KL_divergence(g0), xokdepp::epsilon);
    EXPECT_NEAR(1.9176436331685363, g0.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(11.539183043990622, g_0.KL_divergence(g1), xokdepp::epsilon);
    EXPECT_NEAR(451.5864415871327, g0.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(10253.039183043991, g_0.KL_divergence(g2), xokdepp::epsilon);
    EXPECT_NEAR(119.68008707208254, g0.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(1217.0391830439905, g_0.KL_divergence(g3), xokdepp::epsilon);
    EXPECT_NEAR(6.6737325570323964, g0.KL_divergence(g_4), xokdepp::epsilon);
    EXPECT_NEAR(28.059183043990622, g_0.KL_divergence(g4), xokdepp::epsilon);

    EXPECT_NEAR(1.9176436331685363, g1.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(11.539183043990622, g_1.KL_divergence(g0), xokdepp::epsilon);
    EXPECT_NEAR(1.3642302945889573, g1.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(10.039183043990622, g_1.KL_divergence(g1), xokdepp::epsilon);
    EXPECT_NEAR(443.38852696978074, g1.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(10104.539183043991, g_1.KL_divergence(g2), xokdepp::epsilon);
    EXPECT_NEAR(106.42531626743963, g1.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(1152.5391830439905, g_1.KL_divergence(g3), xokdepp::epsilon);
    EXPECT_NEAR(5.3085938594128805, g1.KL_divergence(g_4), xokdepp::epsilon);
    EXPECT_NEAR(23.359183043990619, g_1.KL_divergence(g4), xokdepp::epsilon);

    EXPECT_NEAR(451.5864415871327, g2.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(10253.039183043991, g_2.KL_divergence(g0), xokdepp::epsilon);
    EXPECT_NEAR(443.38852696978074, g2.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(10104.539183043991, g_2.KL_divergence(g1), xokdepp::epsilon);
    EXPECT_NEAR(1.3642302945889573, g2.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(10.039183043990622, g_2.KL_divergence(g2), xokdepp::epsilon);
    EXPECT_NEAR(257.09667176616085, g2.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(4690.0391830439912, g_2.KL_divergence(g3), xokdepp::epsilon);
    EXPECT_NEAR(446.15386240161234, g2.KL_divergence(g_4), xokdepp::epsilon);
    EXPECT_NEAR(10234.859183043991, g_2.KL_divergence(g4), xokdepp::epsilon);

    EXPECT_NEAR(119.68008707208254, g3.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(1217.0391830439905, g_3.KL_divergence(g0), xokdepp::epsilon);
    EXPECT_NEAR(106.42531626743963, g3.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(1152.5391830439905, g_3.KL_divergence(g1), xokdepp::epsilon);
    EXPECT_NEAR(257.09667176616085, g3.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(4690.0391830439912, g_3.KL_divergence(g2), xokdepp::epsilon);
    EXPECT_NEAR(1.3642302945889573, g3.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(10.039183043990622, g_3.KL_divergence(g3), xokdepp::epsilon);
    EXPECT_NEAR(112.40135404040835, g3.KL_divergence(g_4), xokdepp::epsilon);
    EXPECT_NEAR(1196.4591830439906, g_3.KL_divergence(g4), xokdepp::epsilon);

    EXPECT_NEAR(6.6737325570323964, g4.KL_divergence(g_0), xokdepp::epsilon);
    EXPECT_NEAR(28.059183043990622, g_4.KL_divergence(g0), xokdepp::epsilon);
    EXPECT_NEAR(5.3085938594128805, g4.KL_divergence(g_1), xokdepp::epsilon);
    EXPECT_NEAR(23.359183043990619, g_4.KL_divergence(g1), xokdepp::epsilon);
    EXPECT_NEAR(446.15386240161234, g4.KL_divergence(g_2), xokdepp::epsilon);
    EXPECT_NEAR(10234.859183043991, g_4.KL_divergence(g2), xokdepp::epsilon);
    EXPECT_NEAR(112.40135404040835, g4.KL_divergence(g_3), xokdepp::epsilon);
    EXPECT_NEAR(1196.4591830439906, g_4.KL_divergence(g3), xokdepp::epsilon);
    EXPECT_NEAR(1.3642302945889573, g4.KL_divergence(g_4), xokdepp::epsilon);
    EXPECT_NEAR(10.039183043990622, g_4.KL_divergence(g4), xokdepp::epsilon);
  }

}
