// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <gtest/gtest.h>
#include "kde/gaussian.h"

namespace xokdepp {

  typedef gaussian<matrix_type> gaussian_type;

  class gaussian_full_test: public testing::Test {

  protected:
    virtual void SetUp() {
      //initialization will depend on the test
    }

  };

  static constexpr size_t dims = 3;

  TEST_F(gaussian_full_test,EmptyConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const matrix_type c1 = matrix_type::Zero(dims, dims);
    const gaussian_type g1(dims);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(gaussian_full_test,EmptyConstructor2) {
    //test attributions after empty constructor
    const vector_type v1 = vector_type::Constant(dims, 1);
    const matrix_type c1 = matrix_type::Zero(dims, dims);
    gaussian_type g1(dims);
    g1.mean(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(gaussian_full_test,MeanConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const matrix_type c1 = matrix_type::Zero(dims, dims);
    const gaussian_type g1(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance());
  }

  TEST_F(gaussian_full_test,MeanCovarianceConstructor) {
    const vector_type v1 = vector_type::Zero(dims);
    const matrix_type c1 = matrix_type::Zero(dims, dims);
    const gaussian_type g1(v1, c1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c1, g1.covariance())<< "Ref: " << std::endl << c1 << std::endl << "Actual:" << std::endl << g1.covariance();
  }

  TEST_F(gaussian_full_test,MeanCovarianceConstructor2) {
    //in this test covariance is diagonal, such that cov is invertible
    const vector_type v1 = vector_type::Constant(dims, 1);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;
    const gaussian_type g1(v1, c_diag_1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(c_diag_1, g1.covariance())<< "Ref: " << std::endl << c_diag_1 << std::endl << "Actual:" << std::endl << g1.covariance();

  }

  TEST_F(gaussian_full_test,moment_match_constructor) {
    const vector_type v0 = vector_type::Zero(dims);
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v_ref(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v_ref << 0.5, 0.5, 0.5;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;
    matrix_type c_diag_2(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;

    matrix_type c_ref(dims, dims);
    c_ref << 1.25, 0.25, 0.25, 0.25, 1.25, 0.25, 0.25, 0.25, 1.25;

    vector_type v_ref2(dims), v_ref3(dims);
    v_ref2 << 6.6, 6, 23.5;
    v_ref3 << 0.733333333333333, 4.33333333333333, 48;
    matrix_type c_ref2(dims, dims), c_ref3(dims, dims);
    c_ref2 << 42.11, 0.500000000000007, 152.9, 0.500000000000007, 2.2, 1, 152.9, 1, 564.25;
    c_ref3 << 1.34222222222222, 0.222222222222222, 16.1333333333333, 0.222222222222222, 8.15555555555556, 79.6666666666666, 16.1333333333333, 79.6666666666666, 4528.33333333333;

    //first case
    gaussian_type g1(v0, c_diag_1), g2(v1, c_diag_1);
    mixture<gaussian_type> m1(dims);
    m1.add(g1);
    m1.add(g2);
    m1.normalize_weights();

    gaussian_type g_m(m1, c_diag_2);

    ASSERT_EQ(v_ref, g_m.mean())<< "Ref: " << std::endl << v_ref << std::endl << "Actual:" << std::endl << g_m.mean();
    ASSERT_EQ(c_ref, g_m.covariance())<< "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl << g_m.covariance();

    //second case
    gaussian_type g3(v3, c_diag_1), g4(v4, c_diag_2);
    mixture<gaussian_type> m2(dims);
    m2.add(g3);
    m2.add(g4);
    m2.normalize_weights();

    gaussian_type g_m2(m2, c_diag_1);

    ASSERT_TRUE(allclose(v_ref2, g_m2.mean(), xokdepp::epsilon12, xokdepp::epsilon));
    ASSERT_TRUE(allclose(c_ref2, g_m2.covariance(),xokdepp::epsilon12,xokdepp::epsilon))<< "Ref: " << std::endl << c_ref2 << std::endl << "Actual:" << std::endl << g_m2.covariance();

    //third case
    gaussian_type g5(v4, c_diag_1), g6(v2, c_diag_2), g7(v1, c_diag_2);
    mixture<gaussian_type> m3(dims);
    m3.add(g5);
    m3.add(g6);
    m3.add(g7);
    m3.normalize_weights();

    gaussian_type g_m3(m3, c_diag_1);

    ASSERT_TRUE(allclose(v_ref3, g_m3.mean(), xokdepp::epsilon12, xokdepp::epsilon));
    ASSERT_TRUE(allclose(c_ref3, g_m3.covariance(), xokdepp::epsilon12, xokdepp::epsilon));
  }

  TEST_F(gaussian_full_test,Reset) {
    vector_type v0 = vector_type::Zero(dims);
    matrix_type c0 = matrix_type::Zero(dims, dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    matrix_type c1 = matrix_type::Constant(dims, dims, 1);
    gaussian_type g1(v1, c1);
    g1.reset_mean();
    g1.reset_covariance();
    ASSERT_EQ(v0, g1.mean());
    ASSERT_EQ(c0, g1.covariance())<< "Ref: " << std::endl << c0 << std::endl << "Actual:" << std::endl << g1.covariance();

  }

  TEST_F(gaussian_full_test,Probability) {
    //<< "Ref: " << std::endl << -std::numeric_limits<data_type>::infinity() << std::endl << "Actual:" << std::endl << g1.log_likelihood(obs2);
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;
    gaussian_type g1(v0, c_diag_1);

    vector_type obs1(dims), obs2(dims), obs3(dims), obs4(dims);
    obs1 << 1, 1, 1;
    ASSERT_NEAR(0.014167345154413, g1.likelihood(obs1), xokdepp::epsilon);
    ASSERT_NEAR(-4.256815599614018, g1.log_likelihood(obs1), xokdepp::epsilon);
    obs2 << 18, 21, 293;
    ASSERT_EQ(0, g1.likelihood(obs2));
    ASSERT_NEAR(-43309.75681559961, g1.log_likelihood(obs2), xokdepp::epsilon);
    obs3 << 1, 0, 0;
    ASSERT_NEAR(0.038510836890749, g1.likelihood(obs3), xokdepp::epsilon);
    ASSERT_NEAR(-3.256815599614018, g1.log_likelihood(obs3), xokdepp::epsilon);
    obs4 << 0, 0, 0;
    ASSERT_NEAR(0.063493635934241, g1.likelihood(obs4), xokdepp::epsilon);
    ASSERT_NEAR(-2.756815599614018, g1.log_likelihood(obs4), xokdepp::epsilon);
  }

  TEST_F(gaussian_full_test,Copy) {
    vector_type v1 = vector_type::Constant(dims, 1);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;
    gaussian_type g1(v1, c_diag_1);
    gaussian_type g2(dims);

    g2 = g1;

    ASSERT_EQ(g1.mean(), g2.mean());
    ASSERT_EQ(v1, g2.mean());
    ASSERT_EQ(g1.covariance(), g2.covariance());
    ASSERT_EQ(c_diag_1, g1.covariance());

  }

  TEST_F(gaussian_full_test,modify_gaussian) {
    //this test initializes a gaussian, and then changes its mean and covariance.
    //the data used to initialize the gaussian should be untouched
    vector_type v_ref = vector_type::Zero(dims);
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;
    gaussian_type g1(v0, c_diag_1);

    ASSERT_EQ(v0, g1.mean());
    g1.mean(v1);
    ASSERT_EQ(v1, g1.mean());
    ASSERT_EQ(v_ref, v0);
  }

  TEST_F(gaussian_full_test,KL_divergence) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;
    matrix_type c_diag_2(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;

    gaussian_type g0(v0, c_diag_1);
    gaussian_type g1(v1, c_diag_1);
    gaussian_type g2(v2, c_diag_1);
    gaussian_type g3(v3, c_diag_1);
    gaussian_type g4(v4, c_diag_1);

    ASSERT_EQ(0, g0.KL_divergence(g0));
    ASSERT_EQ(1.5, g0.KL_divergence(g1));
    ASSERT_EQ(10243, g0.KL_divergence(g2));
    ASSERT_EQ(1207, g0.KL_divergence(g3));
    ASSERT_EQ(18.02, g0.KL_divergence(g4));

    ASSERT_EQ(1.5, g1.KL_divergence(g0));
    ASSERT_EQ(0, g1.KL_divergence(g1));
    ASSERT_EQ(10094.5, g1.KL_divergence(g2));
    ASSERT_EQ(1142.5, g1.KL_divergence(g3));
    ASSERT_EQ(13.32, g1.KL_divergence(g4));

    ASSERT_EQ(10243, g2.KL_divergence(g0));
    ASSERT_EQ(10094.5, g2.KL_divergence(g1));
    ASSERT_EQ(0, g2.KL_divergence(g2));
    ASSERT_EQ(4680, g2.KL_divergence(g3));
    ASSERT_NEAR(10224.819999999999709, g2.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_EQ(1207, g3.KL_divergence(g0));
    ASSERT_EQ(1142.5, g3.KL_divergence(g1));
    ASSERT_EQ(4680, g3.KL_divergence(g2));
    ASSERT_EQ(0, g3.KL_divergence(g3));
    ASSERT_NEAR(1186.420000000000073, g3.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_EQ(18.02, g4.KL_divergence(g0));
    ASSERT_EQ(13.32, g4.KL_divergence(g1));
    ASSERT_NEAR(10224.819999999999709, g4.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(1186.420000000000073, g4.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_EQ(0, g4.KL_divergence(g4));

    //gaussians with non diag-only covariance
    gaussian_type g_0(v0, c_diag_2);
    gaussian_type g_1(v1, c_diag_2);
    gaussian_type g_2(v2, c_diag_2);
    gaussian_type g_3(v3, c_diag_2);
    gaussian_type g_4(v4, c_diag_2);

    ASSERT_EQ(0, g_0.KL_divergence(g_0));
    ASSERT_NEAR(1.851063829787233, g_0.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(4128.757683215127145, g_0.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(74.856973995272000, g_0.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(8.756973995271867, g_0.KL_divergence(g_4), xokdepp::epsilon);

    ASSERT_NEAR(1.851063829787233, g_1.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_EQ(0, g_1.KL_divergence(g_1));
    ASSERT_NEAR(4284.360520094560343, g_1.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(60.757683215130101, g_1.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(13.558392434988177, g_1.KL_divergence(g_4), xokdepp::epsilon);

    ASSERT_NEAR(4128.757683215127145, g_2.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(4284.360520094560343, g_2.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_EQ(0, g_2.KL_divergence(g_2));
    ASSERT_NEAR(4536.851063829784835, g_2.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(3933.278250591014057, g_2.KL_divergence(g_4), xokdepp::epsilon);

    ASSERT_NEAR(74.856973995272000, g_3.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(60.757683215130101, g_3.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(4536.851063829784835, g_3.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_EQ(0, g_3.KL_divergence(g_3));
    ASSERT_NEAR(98.327895981087622, g_3.KL_divergence(g_4), xokdepp::epsilon);

    ASSERT_NEAR(8.756973995271867, g_4.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(13.558392434988177, g_4.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(3933.278250591014057, g_4.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(98.327895981087622, g_4.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_EQ(0, g_4.KL_divergence(g_4));

    //now mixed comparison
    ASSERT_NEAR(4.444743144728686, g0.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(11.282325413190932, g_0.KL_divergence(g0), xokdepp::epsilon);
    ASSERT_NEAR(6.295806974515919, g0.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(12.782325413190932, g_0.KL_divergence(g1), xokdepp::epsilon);
    ASSERT_NEAR(4133.202426359855963, g0.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(10254.282325413190847, g_0.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(79.301717140000690, g0.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(1218.282325413190847, g_0.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_NEAR(13.201717140000554, g0.KL_divergence(g_4), xokdepp::epsilon);
    ASSERT_NEAR(29.302325413190932, g_0.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_NEAR(6.295806974515919, g1.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(12.782325413190932, g_1.KL_divergence(g0), xokdepp::epsilon);
    ASSERT_NEAR(4.444743144728686, g1.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(11.282325413190932, g_1.KL_divergence(g1), xokdepp::epsilon);
    ASSERT_NEAR(4288.805263239289161, g1.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(10105.782325413190847, g_1.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(65.202426359858791, g1.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(1153.782325413190847, g_1.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_NEAR(18.003135579716862, g1.KL_divergence(g_4), xokdepp::epsilon);
    ASSERT_NEAR(24.602325413190933, g_1.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_NEAR(4133.202426359855963, g2.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(10254.282325413190847, g_2.KL_divergence(g0), xokdepp::epsilon);
    ASSERT_NEAR(4288.805263239289161, g2.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(10105.782325413190847, g_2.KL_divergence(g1), xokdepp::epsilon);
    ASSERT_NEAR(4.444743144728686, g2.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(11.282325413190932, g_2.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(4541.295806974513653, g2.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(4691.282325413190847, g_2.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_NEAR(3937.722993735742875, g2.KL_divergence(g_4), xokdepp::epsilon);
    ASSERT_NEAR(10236.102325413190556, g_2.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_NEAR(79.301717140000690, g3.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(1218.282325413190847, g_3.KL_divergence(g0), xokdepp::epsilon);
    ASSERT_NEAR(65.202426359858791, g3.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(1153.782325413190847, g_3.KL_divergence(g1), xokdepp::epsilon);
    ASSERT_NEAR(4541.295806974513653, g3.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(4691.282325413190847, g_3.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(4.444743144728686, g3.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(11.282325413190932, g_3.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_NEAR(102.772639125816312, g3.KL_divergence(g_4), xokdepp::epsilon);
    ASSERT_NEAR(1197.702325413190920, g_3.KL_divergence(g4), xokdepp::epsilon);

    ASSERT_NEAR(13.201717140000554, g4.KL_divergence(g_0), xokdepp::epsilon);
    ASSERT_NEAR(29.302325413190932, g_4.KL_divergence(g0), xokdepp::epsilon);
    ASSERT_NEAR(18.003135579716862, g4.KL_divergence(g_1), xokdepp::epsilon);
    ASSERT_NEAR(24.602325413190933, g_4.KL_divergence(g1), xokdepp::epsilon);
    ASSERT_NEAR(3937.722993735742875, g4.KL_divergence(g_2), xokdepp::epsilon);
    ASSERT_NEAR(10236.102325413190556, g_4.KL_divergence(g2), xokdepp::epsilon);
    ASSERT_NEAR(102.772639125816312, g4.KL_divergence(g_3), xokdepp::epsilon);
    ASSERT_NEAR(1197.702325413190920, g_4.KL_divergence(g3), xokdepp::epsilon);
    ASSERT_NEAR(4.444743144728686, g4.KL_divergence(g_4), xokdepp::epsilon);
    ASSERT_NEAR(11.282325413190932, g_4.KL_divergence(g4), xokdepp::epsilon);
  }

}

