// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <gtest/gtest.h>
#include "kde/basic_types.h"
#include "kde/mixture.h"
#include "kde/gaussian.h"
#include "kde/explanation.h"
#include "kde/hellinger_distance.h"

namespace xokdepp {

  typedef gaussian<matrix_type> gaussian_type;
  typedef explanation<gaussian_type, matrix_type> explanation_type;

  class mixture_gaussian_test: public testing::Test {

  protected:
    virtual void SetUp() {
      //initialization will depend on the test
    }

  };

  static constexpr size_t dims = 3;

  TEST_F(mixture_gaussian_test,mix_size_and_weight_normalizing) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    matrix_type c1 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    gaussian_type g0(v0, c_diag_1), g1(v1, c_diag_1);

    mixture<gaussian_type> m(dims);
    m.add(g0);

    ASSERT_EQ(1, m.size());
    ASSERT_EQ(1, m.weight(0));

    m.reset();
    m.add( { g0, g1 });
    m.normalize_weights();

    ASSERT_EQ(2, m.size());
    ASSERT_EQ(0.5, m.weight(0));
    ASSERT_EQ(0.5, m.weight(1));

    m.reset();
    m.add( { g0, g1, g1 });
    m.normalize_weights();

    ASSERT_EQ(3, m.size());
    ASSERT_NEAR(1.0 / 3, m.weight(0), xokdepp::epsilon);
    ASSERT_NEAR(1.0 / 3, m.weight(1), xokdepp::epsilon);
    ASSERT_NEAR(1.0 / 3, m.weight(2), xokdepp::epsilon);

    m.reset();
    m.add( { g0, g1, g1, g1 });
    m.normalize_weights();

    ASSERT_EQ(4, m.size());
    ASSERT_EQ(1.0 / 4, m.weight(0));
    ASSERT_EQ(1.0 / 4, m.weight(1));
    ASSERT_EQ(1.0 / 4, m.weight(2));
    ASSERT_EQ(1.0 / 4, m.weight(3));
  }

  TEST_F(mixture_gaussian_test,goldberger_k_means) {
    constexpr size_t dims = 3;
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    const gaussian_type g0(v0, c_diag_1);
    const gaussian_type g1(v1, c_diag_1);
    const gaussian_type g2(v2, c_diag_1);
    const gaussian_type g3(v3, c_diag_2);
    const gaussian_type g4(v4, c_diag_2);
    const gaussian_type g5(v5, c_diag_2);
    const gaussian_type g6(v6, c_diag_2);

    mixture<gaussian_type> m(dims);
    m.add( { g0, g1, g2, g3, g4 });
    m.normalize_weights();

    ASSERT_NEAR(1.0 / 5, m.weight(0), xokdepp::epsilon);

    int mapping[5];
    mixture<gaussian_type> m_res(dims, 2);
    goldberger_k_means(m, c_diag_1, m_res, mapping);

    int ref_mapping[] = { 1, 1, 0, 1, 1 };

    bool equal = false;
    for (int i = 0; i < 5; i++) {
      equal = (ref_mapping[i] == mapping[i]);
      if (!equal)
        break;
    }
    ASSERT_TRUE(equal);

    mixture<gaussian_type> m2(dims);
    m2.add( { g0, g1, g2, g3, g4, g5, g6 });
    m2.normalize_weights();

    ASSERT_NEAR(1.0 / 7, m2.weight(0), xokdepp::epsilon);

    int mapping2[7];
    mixture<gaussian_type> m_res2(dims, 2);
    goldberger_k_means(m2, c_diag_1, m_res2, mapping2);

    int ref_mapping2[] = { 1, 1, 0, 1, 1, 0, 0 };

    bool equal2 = false;
    for (int i = 0; i < 7; i++) {
      equal2 = (ref_mapping2[i] == mapping2[i]);
      if (!equal2)
        break;
    }

    ASSERT_TRUE(equal2);

    mixture<gaussian_type> m3(dims);
    m3.add( { g3, g3, g3, g3, g3, g3, g3 });
    m3.normalize_weights();

    ASSERT_NEAR(1.0 / 7, m3.weight(0), xokdepp::epsilon);

    int mapping3[7];
    mixture<gaussian_type> m_res3(dims, 2);
    goldberger_k_means(m3, c_diag_1, m_res3, mapping3);

    int ref_mapping3[] = { 0, 1, 1, 1, 1, 1, 1 };

//    for (int i = 0; i < 7; i++) {
//      std::cout << "R#" << i << ": " << ref_mapping3[i] << std::endl;
//      std::cout << "A#" << i << ": " << mapping3[i] << std::endl;
//    }

    bool equal3 = false;
    for (int i = 0; i < 7; i++) {
      equal3 = (ref_mapping3[i] == mapping3[i]);
      if (!equal3)
        break;
    }

    ASSERT_TRUE(equal3);
  }

  TEST_F(mixture_gaussian_test,goldberger_k_means_explanation) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    explanation_type g0(v0, c_diag_1);
    explanation_type g1(v1, c_diag_1);
    explanation_type g2(v2, c_diag_1);
    explanation_type g3(v3, c_diag_2);
    explanation_type g4(v4, c_diag_2);
    explanation_type g5(v5, c_diag_2);
    explanation_type g6(v6, c_diag_2);

    mixture<explanation_type> m(dims);
    m.add( { g0, g1, g2, g3, g4 });
    m.normalize_weights();

    ASSERT_NEAR(1.0 / 5, m.weight(0), xokdepp::epsilon);

    int mapping[5];
    mixture<explanation_type> m_res(dims, 2);
    goldberger_k_means(m, c_diag_1, m_res, mapping);

    int ref_mapping[] = { 1, 1, 0, 1, 1 };

    bool equal = false;
    for (int i = 0; i < 5; i++) {
      equal = (ref_mapping[i] == mapping[i]);
      if (!equal)
        break;
    }
    ASSERT_TRUE(equal);

    mixture<explanation_type> m2(dims);
    m2.add( { g0, g1, g2, g3, g4, g5, g6 });
    m2.normalize_weights();

    ASSERT_NEAR(1.0 / 7, m2.weight(0), xokdepp::epsilon);

    int mapping2[7];
    mixture<explanation_type> m_res2(dims, 2);
    goldberger_k_means(m2, c_diag_1, m_res2, mapping2);

    int ref_mapping2[] = { 1, 1, 0, 1, 1, 0, 0 };

    bool equal2 = false;
    for (int i = 0; i < 7; i++) {
      equal2 = (ref_mapping2[i] == mapping2[i]);
      if (!equal2)
        break;
    }

    ASSERT_TRUE(equal2);

    mixture<explanation_type> m3(dims);
    m3.add( { g3, g3, g3, g3, g3, g3, g3 });
    m3.normalize_weights();

    ASSERT_NEAR(1.0 / 7, m3.weight(0), xokdepp::epsilon);

    int mapping3[7];
    mixture<explanation_type> m_res3(dims, 2);
    goldberger_k_means(m3, c_diag_1, m_res3, mapping3);

    //int ref_mapping3[] = { 1, 0, 0, 0, 0, 0, 0 };
    int ref_mapping3[] = { 0, 1, 1, 1, 1, 1, 1 };

//    for (int i = 0; i < 7; i++) {
//      std::cout << "R#" << i << ": " << ref_mapping3[i] << std::endl;
//      std::cout << "A#" << i << ": " << mapping3[i] << std::endl;
//    }

    bool equal3 = false;
    for (int i = 0; i < 7; i++) {
      equal3 = (ref_mapping3[i] == mapping3[i]);
      if (!equal3)
        break;
    }

    ASSERT_TRUE(equal3);

    vector_type vx1(dims), vx2(dims), vx3(dims), vx4(dims), vx5(dims);
    vx1 << 1, 0, 143;
    vx2 << 1, 0, 143;
    vx3 << 13, 0, 47;
    vx4 << 0.2, 0, 0;
    vx5 << 1, 0, 255;

    explanation_type gx1(vx1, c_diag_3);
    explanation_type gx2(vx2, c_diag_3);
    explanation_type gx3(vx3, c_diag_3);
    explanation_type gx4(vx4, c_diag_3);
    explanation_type gx5(vx5, c_diag_3);

    mixture<explanation_type> mx1(dims);
    mx1.add( { gx1, gx2, gx3, gx4, gx5 });
    mx1.normalize_weights();

    int mappingx1[5];
    mixture<explanation_type> m_resx1(dims, 2);
    goldberger_k_means(mx1, c_diag_1, m_resx1, mappingx1);

    int ref_mappingx1[] = { 0, 0, 1, 1, 0 };

    bool equalx1 = false;
    for (int i = 0; i < 5; i++) {
      equalx1 = (ref_mappingx1[i] == mappingx1[i]);
      if (!equal3)
        break;
    }

    ASSERT_TRUE(equalx1);
  }

  TEST_F(mixture_gaussian_test,hellinger_distance) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    explanation_type g0(v0, c_diag_1);
    explanation_type g1(v1, c_diag_1);
    explanation_type g2(v2, c_diag_1);
    explanation_type g3(v3, c_diag_2);
    explanation_type g4(v4, c_diag_2);
    explanation_type g5(v5, c_diag_2);
    explanation_type g6(v6, c_diag_2);

    vector_type vx1(dims), vx2(dims), vx3(dims), vx4(dims), vx5(dims);
    vx1 << 1, 0, 143;
    vx2 << 1, 0, 143;
    vx3 << 13, 0, 47;
    vx4 << 0.2, 0, 0;
    vx5 << 1, 0, 255;

    explanation_type gx1(v1, c_diag_3);
    explanation_type gx2(v2, c_diag_3);
    explanation_type gx3(v3, c_diag_3);
    explanation_type gx4(v4, c_diag_3);
    explanation_type gx5(v5, c_diag_3);

    mixture<explanation_type> m(dims);
    m.add( { g0, g1, g2, g3, g4 });
    m.normalize_weights();

    //minor tweak of one of the covariances to match the setup in matlab
    mixture<explanation_type> m_2(dims);
    m_2.add(explanation_type(v0, c_diag_1));
    m_2.add(explanation_type(v1, c_diag_1));
    m_2.add(explanation_type(v2, c_diag_2));
    m_2.add(explanation_type(v3, c_diag_2));
    m_2.add(explanation_type(v4, c_diag_2));
    m_2.normalize_weights();

    mixture<explanation_type> m2(dims);
    m2.add( { g0, g1, g2, g3, g4, g5, g6 });
    m2.normalize_weights();

    mixture<explanation_type> m3(dims);
    m3.add( { g3, g3, g3, g3, g3, g3, g3 });
    m3.normalize_weights();

    mixture<explanation_type> m4(dims);
    m4.add(explanation_type(v3, c_diag_2));
    m4.add(explanation_type(v2, c_diag_2));
    m4.add(explanation_type(v6, c_diag_1));
    m4.add(explanation_type(v2, c_diag_1));
    m4.add(explanation_type(v4, c_diag_2));
    m4.add(explanation_type(v5, c_diag_2));
    m4.add(explanation_type(v6, c_diag_1));
    m4.normalize_weights();

    mixture<explanation_type> mx1(dims);
    mx1.add( { gx1, gx2, gx3, gx4, gx5 });
    mx1.normalize_weights();

    data_type dist0 = hellinger_distance(m, m);
    data_type dist1 = hellinger_distance(m3, m2);
    data_type dist2 = hellinger_distance(m2, m3);
    data_type dist3 = hellinger_distance(m2, m2);
    data_type dist4 = hellinger_distance(m3, m3);

    ASSERT_EQ(0, dist0);
    ASSERT_NEAR(0.788692289161699, dist1, xokdepp::epsilon);
    ASSERT_NEAR(0.7886922891617, dist2, xokdepp::epsilon);
    ASSERT_EQ(0, dist3);
    ASSERT_EQ(0, dist4);

    data_type dist5 = hellinger_distance(mx1, mx1);
    data_type dist6 = hellinger_distance(m_2, mx1);

    ASSERT_EQ(0, dist5); //not validated by oKDE should be this!
    //	ASSERT_NEAR( 0.99209416503201786, dist6 ,hdkj::epsilon); //not validated by oKDE, ref got by running once. makes some sense
    ASSERT_NEAR(0.728855608806175, dist6, xokdepp::epsilon);
    data_type dist7 = hellinger_distance(m4, m4);
    data_type dist8 = hellinger_distance(m4, m2);
    data_type dist9 = hellinger_distance(m4, m3);
    ASSERT_EQ(0, dist7);
    ASSERT_NEAR(0.527675693466034, dist8, xokdepp::epsilon);
    ASSERT_NEAR(0.78869228916655, dist9, xokdepp::epsilon);

  }


//  TEST_F(mixture_test,mix_probability) {
//    vector_type v0 = vector_type::Zero(dims);
//    vector_type v1 = vector_type::Constant(dims, 1);
//    matrix_type c1 = matrix_type::Zero(dims, dims);
//    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
//    c_diag_1.diagonal() = v1;
//
//    gaussian_type g0(v0, c_diag_1), g1(v1, c_diag_1);
//
//    mixture<gaussian_type> m(dims);
//    m.add(g0);
//    m.add(g1);
//
//    ASSERT_EQ(0.077661, m.likelihood(v1)) << std::cout << m.likelihood(v1) << std::endl ;
//
//  }

}
