// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <gtest/gtest.h>
#include "kde/okde.h"
#include "kde/explanation.h"

namespace xokdepp {

  typedef gaussian<vector_type> gaussian_type;
  typedef explanation<gaussian_type, vector_type> explanation_type;
  typedef oKDE<gaussian_type, vector_type> kde_type;

  class okde_gaussian_diagonal_test: public testing::Test {

  protected:
    virtual void SetUp() {
      //initialization will depend on the test
    }

  };

  static constexpr size_t dims = 3;

  TEST_F(okde_gaussian_diagonal_test,contructor_test1) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    const vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims), c_diag_3(dims);
    c_diag_2 << 1.3, 3.4, 23;
    c_diag_3 << 1.3, 0, 23;

    kde_type m(dims);
    m.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m.add(explanation_type(v4, c_diag_2), 1 / 5.0);

    ASSERT_TRUE(allclose(m.mean(0), v0, xokdepp::epsilon12, xokdepp::epsilon));
    ASSERT_EQ(1 / 5.0, m.weight(0));
    ASSERT_EQ(1 / 5.0, m.weight(1));
    ASSERT_EQ(1 / 5.0, m.weight(2));
    ASSERT_EQ(1 / 5.0, m.weight(3));
    ASSERT_EQ(1 / 5.0, m.weight(4));
  }

  TEST_F(okde_gaussian_diagonal_test,moment_matching2) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;

    const vector_type c_0 = vector_type::Zero(dims);
    const vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims), c_diag_3(dims);
    c_diag_2 << 1.3, 3.4, 23;
    c_diag_3 << 1.3, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);

    explanation_type pdf0(m0);
    //std::cout << pdf0 << std::endl;

    //ref
    vector_type v_ref(dims);
    v_ref << 3.04, 3.8000000000000007, 38.200000000000003;
    vector_type c_ref(dims, dims);
    c_ref << 26.146400000000007, 9.7999999999999936, 3086.7599999999998;

    EXPECT_TRUE(allclose(v_ref, pdf0.mean())) << "Ref: " << std::endl << v_ref << std::endl << "Actual:" << std::endl
        << pdf0.mean();
    EXPECT_TRUE(allclose(c_ref, pdf0.covariance())) << "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl
        << pdf0.covariance();

  }

  TEST_F(okde_gaussian_diagonal_test,optimal_bandwidth) {
    const vector_type v0 = vector_type::Zero(dims);
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    const vector_type c_0 = vector_type::Zero(dims);
    const vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims), c_diag_3(dims);
    c_diag_2 << 1.3, 3.4, 23;
    c_diag_3 << 1.3, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);
    m0.estimate_bandwidth();

    std::cout << m0.optimal_bandwidth() << std::endl;

    std::cout << "DEGENERATE CASE 1D" << std::endl;

    kde_type m(dims);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx2, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.debug_set_Nt(5);
    m.estimate_bandwidth();

    std::cout << m.optimal_bandwidth() << std::endl;

    std::cout << "DEGENERATE CASE 2D" << std::endl;

    kde_type m2(dims);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.add(explanation_type(vx2, c_0), 1 / 5.0);
    m2.add(explanation_type(vx3, c_0), 1 / 5.0);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.debug_set_Nt(5);
    m2.estimate_bandwidth();

    std::cout << m2.optimal_bandwidth() << std::endl;
  }

  TEST_F(okde_gaussian_diagonal_test,optimal_bandwidth3) {
    const vector_type v0 = vector_type::Zero(dims);
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    vector_type c_0 = vector_type::Zero(dims);
    vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims), c_diag_3(dims);
    c_diag_2 << 1.3, 3.4, 23;
    c_diag_3 << 1.3, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);

    m0.estimate_bandwidth();
    std::cout << "optimal_bandwidth" << std::endl << m0.optimal_bandwidth() << std::endl;
    std::cout << "model" << std::endl << m0.component(0) << std::endl;
    std::cout << "---------" << std::endl;

    m0.estimate_kernel_density();
    std::cout << "optimal_bandwidth" << std::endl << m0.optimal_bandwidth() << std::endl;
    std::cout << "model" << std::endl << m0.component(0) << std::endl;
    std::cout << "---------" << std::endl;

    m0.estimate_kernel_density();
    std::cout << "optimal_bandwidth" << std::endl << m0.optimal_bandwidth() << std::endl;
    std::cout << "model" << std::endl << m0.component(0) << std::endl;
    std::cout << "---------" << std::endl;

  }

  TEST_F(okde_gaussian_diagonal_test,add_sample) {
    const vector_type v0 = vector_type::Zero(dims);
    const vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    const vector_type c_0 = vector_type::Zero(dims);
    const vector_type c_diag_1 = v1;
    vector_type c_diag_2(dims), c_diag_3(dims);
    c_diag_2 << 1.3, 3.4, 23;
    c_diag_3 << 1.3, 0, 23;

    data_type dth = 0.01;

    kde_type m0(dims, dth);
    m0.add_samples( { v0, v1, v2, v3, v4, v5, v6 });
    m0.add_samples( { vx1, vx2, vx1, vx2 });
    m0.add_samples( { v0, v1, v2, v3, v4, v5, v6 });
    m0.add_samples( { v0, v1, v2, v3, v4, v5, v6 });
    m0.add_samples( { vx1, vx1, vx1, vx1, vx1, vx1 });
    m0.add_samples( { vx2, vx2, vx2, vx2, vx2, vx2 });
    m0.add_sample(v0);

    m0.estimate_kernel_density();
    std::cout << "Final Model:" << std::endl << m0 << std::endl;
    std::cout << "optBand" << std::endl << m0.optimal_bandwidth() << std::endl;

    vector_type ref_opt_band(dims, dims);
    ref_opt_band << 0.43272399028518305, 1.1043453961386624, 230.59329156004955; //original 0.446276853841142, 1.13893336170396, 237.815445839765;

    ASSERT_TRUE(
        allclose(m0.optimal_bandwidth(), ref_opt_band, xokdepp::epsilon12,
            xokdepp::epsilon))<< "Ref: " << std::endl << ref_opt_band << std::endl << "Actual:" << std::endl << m0.optimal_bandwidth();
  }

  TEST_F(okde_gaussian_diagonal_test,whitening_constructor2) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    vector_type c_diag_1 = vector_type::Constant(dims, 1);

    vector_type c_diag_2(dims);
    c_diag_2 << 1.3, 3.4, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);
    m0.estimate_bandwidth();

    std::vector<int> ev_ok, ev_ko;
    matrix_type F_trns;

    explanation_type smp(m0, false);

    //test smp correctness
    vector_type smp_ref_m(dims);
    smp_ref_m << 3.04, 3.8000000000000007, 38.200000000000003;
    vector_type smp_ref_cov(dims);
    smp_ref_cov << 26.146400000000007, 9.7999999999999936, 3086.7599999999998;
    ASSERT_TRUE(allclose(smp_ref_m, smp.mean()))<< "Ref: " << std::endl
    << smp_ref_m << std::endl << "Actual:" << std::endl << smp.mean();
    ASSERT_TRUE(allclose(smp_ref_cov, smp.covariance()))<< "Ref: " << std::endl << smp_ref_cov << std::endl << "Actual:" << std::endl << smp.covariance();

    smp.get_whitening_parameters(ev_ok, ev_ko, F_trns);

    //test whitening parameters
    matrix_type ref_F_trns(dims, dims);
    //original on okde_full
    //ref_F_trns << 0.00020562593514063566, 0.00048529923238563229, 0.017983595023472354, 0.18721070620637167, 0.045842716701635734, -0.0033776790279428401, -0.094172812333361888, 0.38389531308129632, -0.0092828896522654455;
    ref_F_trns << 0, 0, 0.017999007922023962, 0.1955663136401237, 0, 0, 0, 0.31943828249834028, 0;
    Eigen::VectorXi ref_ev_ok(dims);
    ref_ev_ok << 0, 1, 2;
    Eigen::VectorXi ref_ev_ko(dims);	//real_ev_ko should have dim 0

    //convert the eig_ok and eig_ko into eigen::vector
    int* ptr_ok = &ev_ok[0];
    Eigen::Map<Eigen::VectorXi> real_ev_ok(ptr_ok, ev_ok.size());
    int* ptr_ko = &ev_ko[0];
    Eigen::Map<Eigen::VectorXi> real_ev_ko(ptr_ko, ev_ko.size());

    EXPECT_TRUE(allclose(ref_F_trns,F_trns)) << "Ref: " << std::endl << ref_F_trns << std::endl << "Actual:" << std::endl << F_trns;
    EXPECT_TRUE(allclose(ref_ev_ok, real_ev_ok)) << "Ref: " << std::endl << ref_ev_ok << std::endl << "Actual:" << std::endl
        << real_ev_ok;
    EXPECT_TRUE(allclose(ref_ev_ko, real_ev_ko)) << "Ref: " << std::endl << ref_ev_ko << std::endl << "Actual:" << std::endl
        << real_ev_ko;

    kde_type whitened_kde(m0, smp.mean(), ev_ok, ev_ko, F_trns);

    //std::cout << "whitened_kde: " << std::endl << whitened_kde << std::endl;
    vector_type ref_m0(dims), ref_m1(dims);
    ref_m0 << -0.68756210262131534, -0.59452159346597599, -1.2138654734936933;
    ref_m1 << -0.66956309469929143, -0.39895527982585233, -0.89442719099535295;

    vector_type ref_cov0(dims), ref_cov1(dims);
    ref_cov0 << 0.15803526648215194, 0.19595748522676207, 0.25975211852146429;
    ref_cov1 << 0.15803526648215194, 0.19595748522676207, 0.25975211852146429;

    vector_type ref_base_cov0(dims), ref_base_cov1(dims);
    ref_base_cov0 << 0.00032396428617708132, 0.03824618303078723, 0.10204081632548945;
    ref_base_cov1 << 0.00032396428617708132, 0.03824618303078723, 0.10204081632548945;

    EXPECT_TRUE(allclose(ref_m0, whitened_kde.component(0).mean())) << "Ref: " << std::endl << ref_m0 << std::endl << "Actual:"
        << std::endl << whitened_kde.component(0).mean();
    EXPECT_TRUE(allclose(ref_m1, whitened_kde.component(1).mean())) << "Ref: " << std::endl << ref_m1 << std::endl << "Actual:"
        << std::endl << whitened_kde.component(1).mean();

    EXPECT_TRUE(allclose(ref_cov0, whitened_kde.component(0).covariance())) << "Ref: " << std::endl << ref_cov0 << std::endl
        << "Actual:" << std::endl << whitened_kde.component(0).covariance();
    EXPECT_TRUE(allclose(ref_base_cov0, whitened_kde.component(0).base_covariance())) << "Ref: " << std::endl << ref_base_cov0
        << std::endl << "Actual:" << std::endl << whitened_kde.component(0).base_covariance();

    EXPECT_TRUE(allclose(ref_cov1, whitened_kde.component(1).covariance())) << "Ref: " << std::endl << ref_cov1 << std::endl
        << "Actual:" << std::endl << whitened_kde.component(1).covariance();
    EXPECT_TRUE(allclose(ref_base_cov1, whitened_kde.component(1).base_covariance())) << "Ref: " << std::endl << ref_base_cov1
        << std::endl << "Actual:" << std::endl << whitened_kde.component(1).base_covariance();

  }

}

