// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <gtest/gtest.h>
#include "kde/okde.h"
#include "kde/explanation.h"

namespace xokdepp {

  typedef gaussian<matrix_type> gaussian_type;
  typedef explanation<gaussian_type, matrix_type> explanation_type;
  typedef oKDE<gaussian_type, matrix_type> kde_type;

  class okde_gaussian_full_test: public testing::Test {

  protected:
    virtual void SetUp() {
      //initialization will depend on the test
    }

  };

  static constexpr size_t dims = 3;

  TEST_F(okde_gaussian_full_test,contructor_test1) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    kde_type m(dims);
    m.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m.add(explanation_type(v4, c_diag_2), 1 / 5.0);

    ASSERT_TRUE(allclose(m.mean(0), v0, xokdepp::epsilon12, xokdepp::epsilon));
    ASSERT_EQ(1 / 5.0, m.weight(0));
    ASSERT_EQ(1 / 5.0, m.weight(1));
    ASSERT_EQ(1 / 5.0, m.weight(2));
    ASSERT_EQ(1 / 5.0, m.weight(3));
    ASSERT_EQ(1 / 5.0, m.weight(4));
  }

  TEST_F(okde_gaussian_full_test,moment_matching) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);

    explanation_type pdf0(m0);
    std::cout << pdf0 << std::endl;

    //ref
    vector_type v_ref(dims);
    v_ref << 3.04, 3.8000000000000007, 38.200000000000003;
    matrix_type c_ref(dims, dims);
    c_ref << 26.146400000000007, 6.2880000000000003, 37.871999999999986, 6.2880000000000003, 9.7999999999999936, 84.239999999999981, 37.871999999999986, 84.239999999999981, 3086.7599999999998;

    EXPECT_TRUE(allclose(v_ref, pdf0.mean())) << "Ref: " << std::endl << v_ref << std::endl << "Actual:" << std::endl
        << pdf0.mean();
    EXPECT_TRUE(allclose(c_ref, pdf0.covariance())) << "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl
        << pdf0.covariance();

  }

  TEST_F(okde_gaussian_full_test,moment_matching2) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims);
    c_diag_2 << 1.3, 0, 0, 0, 3.4, 0, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);

    explanation_type pdf0(m0);
    std::cout << pdf0 << std::endl;

    //ref
    vector_type v_ref(dims);
    v_ref << 3.04, 3.8000000000000007, 38.200000000000003;
    matrix_type c_ref(dims, dims);
    c_ref << 26.146400000000007, 5.6879999999999953, 34.871999999999986, 5.6879999999999953, 9.7999999999999936, 83.039999999999992, 34.871999999999986, 96.839999999999975, 3086.7599999999998;

    EXPECT_TRUE(allclose(v_ref, pdf0.mean())) << "Ref: " << std::endl << v_ref << std::endl << "Actual:" << std::endl
        << pdf0.mean();
    EXPECT_TRUE(allclose(c_ref, pdf0.covariance())) << "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl
        << pdf0.covariance();

  }

  TEST_F(okde_gaussian_full_test,optimal_bandwidth) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);

    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);
    m0.estimate_bandwidth();

    std::cout << m0.optimal_bandwidth() << std::endl;

    std::cout << "DEGENERATE CASE 1D" << std::endl;

    kde_type m(dims);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx2, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.debug_set_Nt(5);
    m.estimate_bandwidth();

    std::cout << m.optimal_bandwidth() << std::endl;

    std::cout << "DEGENERATE CASE 2D" << std::endl;

    kde_type m2(dims);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.add(explanation_type(vx2, c_0), 1 / 5.0);
    m2.add(explanation_type(vx3, c_0), 1 / 5.0);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.debug_set_Nt(5);
    m2.estimate_bandwidth();

    std::cout << m2.optimal_bandwidth() << std::endl;

    //ref
    matrix_type c_ref0(dims, dims), c_ref1(dims, dims), c_ref2(dims, dims);
    c_ref0 << 4.0551660829935878, 0.97523499716458051, 5.8737436088767492, 0.9752349971645804, 1.5199273174637948, 13.065171145219681, 5.8737436088767492, 13.065171145219681, 478.73988229130094;
    c_ref1 << 0.001599999999999999, 0, 0, 0, 0.031912901395322484, 0, 0, 0, 0.001599999999999999;
    c_ref2 << 0.002, 0, 0, 0, 0.032335418197904982, 0.016167709098952491, 0, 0.016167709098952494, 0.021556945465269983;

    EXPECT_TRUE(allclose(c_ref0, m0.optimal_bandwidth())) << "Ref: " << std::endl << c_ref0 << std::endl << "Actual:" << std::endl
        << m0.optimal_bandwidth();
    EXPECT_TRUE(allclose(c_ref1, m.optimal_bandwidth())) << "Ref: " << std::endl << c_ref1 << std::endl << "Actual:" << std::endl
        << m.optimal_bandwidth();
    EXPECT_TRUE(allclose(c_ref2, m2.optimal_bandwidth())) << "Ref: " << std::endl << c_ref2 << std::endl << "Actual:" << std::endl
        << m2.optimal_bandwidth();

  }

  TEST_F(okde_gaussian_full_test,optimal_bandwidth2) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = vector_type::Constant(dims, 1);

    matrix_type c_diag_2(dims, dims);

    c_diag_2 << 1.3, 0, 0, 0, 3.4, 0, 0, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);
    m0.estimate_bandwidth();

    std::cout << m0.optimal_bandwidth() << std::endl;

    std::cout << "DEGENERATE CASE 1D" << std::endl;

    kde_type m(dims);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx2, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.add(explanation_type(vx1, c_0), 1 / 5.0);
    m.debug_set_Nt(5);
    m.estimate_bandwidth();

    std::cout << m.optimal_bandwidth() << std::endl;

    std::cout << "DEGENERATE CASE 2D" << std::endl;

    kde_type m2(dims);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.add(explanation_type(vx2, c_0), 1 / 5.0);
    m2.add(explanation_type(vx3, c_0), 1 / 5.0);
    m2.add(explanation_type(vx1, c_0), 1 / 5.0);
    m2.debug_set_Nt(5);
    m2.estimate_bandwidth();

    std::cout << m2.optimal_bandwidth() << std::endl;

    //ref
    matrix_type c_ref0(dims, dims), c_ref1(dims, dims), c_ref2(dims, dims);
    c_ref0 << 4.1177599351312191, 0.8957951576900185, 5.4919424646565167, 0.89579515769001827, 1.5433882815334417, 13.077853356993442, 5.4919424646565167, 13.077853356993442, 486.12951141899572;
    c_ref1 << 0.001599999999999999, 0, 0, 0, 0.031912901395322484, 0, 0, 0, 0.001599999999999999;
    c_ref2 << 0.002, 0, 0, 0, 0.032335418197904982, 0.016167709098952491, 0, 0.016167709098952494, 0.021556945465269983;

    EXPECT_TRUE(allclose(c_ref0, m0.optimal_bandwidth())) << "Ref: " << std::endl << c_ref0 << std::endl << "Actual:" << std::endl
        << m0.optimal_bandwidth();
    EXPECT_TRUE(allclose(c_ref1, m.optimal_bandwidth())) << "Ref: " << std::endl << c_ref1 << std::endl << "Actual:" << std::endl
        << m.optimal_bandwidth();
    EXPECT_TRUE(allclose(c_ref2, m2.optimal_bandwidth())) << "Ref: " << std::endl << c_ref2 << std::endl << "Actual:" << std::endl
        << m2.optimal_bandwidth();

  }

  TEST_F(okde_gaussian_full_test,optimal_bandwidth3) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);

    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    //ref
    matrix_type c_ref(dims, dims);
    c_ref << 4.0551660829935878, 0.97523499716458051, 5.8737436088767492, 0.9752349971645804, 1.5199273174637948, 13.065171145219681, 5.8737436088767492, 13.065171145219681, 478.73988229130094;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);

    m0.estimate_bandwidth();
//    std::cout << "optimal_bandwidth" << std::endl << m0.optimal_bandwidth() << std::endl;
//    std::cout << "model" << std::endl << m0.component(0) << std::endl;
//    std::cout << "---------" << std::endl;
    EXPECT_TRUE(allclose(c_ref, m0.optimal_bandwidth())) << "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl
        << m0.optimal_bandwidth();

    m0.estimate_kernel_density();
//    std::cout << "optimal_bandwidth" << std::endl << m0.optimal_bandwidth() << std::endl;
//    std::cout << "model" << std::endl << m0.component(0) << std::endl;
//    std::cout << "---------" << std::endl;
    EXPECT_TRUE(allclose(c_ref, m0.optimal_bandwidth())) << "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl
        << m0.optimal_bandwidth();

    m0.estimate_kernel_density();
//    std::cout << "optimal_bandwidth" << std::endl << m0.optimal_bandwidth() << std::endl;
//    std::cout << "model" << std::endl << m0.component(0) << std::endl;
//    std::cout << "---------" << std::endl;
    EXPECT_TRUE(allclose(c_ref, m0.optimal_bandwidth())) << "Ref: " << std::endl << c_ref << std::endl << "Actual:" << std::endl
        << m0.optimal_bandwidth();

  }

  TEST_F(okde_gaussian_full_test,add_sample) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);
    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    data_type dth = 0.01;

    kde_type m0(dims, dth);
    m0.add_samples( { v0, v1, v2, v3, v4, v5, v6 });
    m0.add_samples( { vx1, vx2, vx1, vx2 });
    m0.add_samples( { v0, v1, v2, v3, v4, v5, v6 });
    m0.add_samples( { v0, v1, v2, v3, v4, v5, v6 });
    m0.add_samples( { vx1, vx1, vx1, vx1, vx1, vx1 });
    m0.add_samples( { vx2, vx2, vx2, vx2, vx2, vx2 });
    m0.add_sample(v0);

    m0.estimate_kernel_density();
    std::cout << "Final Model:" << std::endl << m0 << std::endl;
    std::cout << "optBand" << std::endl << m0.optimal_bandwidth() << std::endl;

    matrix_type ref_opt_band(dims, dims);
    ref_opt_band << 0.446276853841142, 0.0869861664266612, 0.434183489386047, 0.0869861664266613, 1.13893336170396, 13.1177901874897, 0.434183489386048, 13.1177901874897, 237.815445839765;

    ASSERT_TRUE(allclose(m0.optimal_bandwidth(), ref_opt_band, xokdepp::epsilon12, xokdepp::epsilon));
  }

  TEST_F(okde_gaussian_full_test,whitening_constructor1) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    matrix_type c_0 = matrix_type::Zero(dims, dims);
    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = v1;

    matrix_type c_diag_2(dims, dims), c_diag_3(dims, dims);

    c_diag_2 << 1.3, 1, 5, 1, 3.4, 2, 5, 2, 23;
    c_diag_3 << 1.3, 0, 5, 0, 0, 0, 5, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);
    m0.estimate_bandwidth();

    std::vector<int> ev_ok, ev_ko;
    matrix_type F_trns;

    explanation_type smp(m0, false);
    smp.get_whitening_parameters(ev_ok, ev_ko, F_trns);
    kde_type whitened_kde(m0, smp.mean(), ev_ok, ev_ko, F_trns);

    //std::cout << "whitened_kde: " << std::endl << whitened_kde << std::endl;
    vector_type ref_m0(dims);
    ref_m0 << -0.6894925083879242, -0.61295219848602445, -0.80865841136614702;
    vector_type ref_m1(dims);
    ref_m1 << -0.67079404821777489, -0.38135476846729005, -0.53163755990508554;

    matrix_type ref_cov0(dims, dims);
    ref_cov0 << 0.15541829735827337, 1.3891340334970526e-17, -2.1648535828561188e-15, 1.9529191631895149e-17, 0.19202045299690662, -4.2846314603911528e-17, -2.1764545461017137e-15, -5.0849081889570158e-17, 0.31934133350905858;

    matrix_type ref_base_cov0(dims, dims);
    ref_base_cov0 << 0.00032367312727230344, 1.3552527156068805e-20, 8.1315162936412833e-20, 1.3552527156068805e-20, 0.036925828766473053, 4.8585809854506667e-18, 8.1315162936412833e-20, 4.8585809854506667e-18, 0.16424670928060858;

    matrix_type ref_cov1(dims, dims);
    ref_cov1 << 0.15541829735827337, 1.3891340334970526e-17, -2.1648535828561188e-15, 1.9529191631895149e-17, 0.19202045299690662, -4.2846314603911528e-17, -2.1764545461017137e-15, -5.0849081889570158e-17, 0.31934133350905858;

    matrix_type ref_base_cov1(dims, dims);
    ref_base_cov1 << 0.00032367312727230344, 1.3552527156068805e-20, 8.1315162936412833e-20, 1.3552527156068805e-20, 0.036925828766473053, 4.8585809854506667e-18, 8.1315162936412833e-20, 4.8585809854506667e-18, 0.16424670928060858;

    EXPECT_TRUE(allclose(ref_m0, whitened_kde.component(0).mean())) << "Ref: " << std::endl << ref_m0 << std::endl << "Actual:"
        << std::endl << whitened_kde.component(0).mean();
    EXPECT_TRUE(allclose(ref_m1, whitened_kde.component(1).mean())) << "Ref: " << std::endl << ref_m1 << std::endl << "Actual:"
        << std::endl << whitened_kde.component(1).mean();

    EXPECT_TRUE(allclose(ref_cov0, whitened_kde.component(0).covariance())) << "Ref: " << std::endl << ref_cov0 << std::endl
        << "Actual:" << std::endl << whitened_kde.component(0).covariance();
    EXPECT_TRUE(allclose(ref_base_cov0, whitened_kde.component(0).base_covariance())) << "Ref: " << std::endl << ref_base_cov0
        << std::endl << "Actual:" << std::endl << whitened_kde.component(0).base_covariance();

    EXPECT_TRUE(allclose(ref_cov1, whitened_kde.component(1).covariance())) << "Ref: " << std::endl << ref_cov1 << std::endl
        << "Actual:" << std::endl << whitened_kde.component(1).covariance();
    EXPECT_TRUE(allclose(ref_base_cov1, whitened_kde.component(1).base_covariance())) << "Ref: " << std::endl << ref_base_cov1
        << std::endl << "Actual:" << std::endl << whitened_kde.component(1).base_covariance();

    //m0.compute_rpfg(const mixture<explanation_type> &mix, const matrix_type &F, data_type N)

  }

  TEST_F(okde_gaussian_full_test,whitening_constructor2) {
    vector_type v0 = vector_type::Zero(dims);
    vector_type v1 = vector_type::Constant(dims, 1);
    vector_type v2(dims), v3(dims), v4(dims), v5(dims), v6(dims);
    v2 << 1, 6, 143;
    v3 << 13, 6, 47;
    v4 << 0.2, 6, 0;
    v5 << 1, 14, 255;
    v6 << 1, 16, 100;

    vector_type vx1(dims), vx2(dims), vx3(dims);
    vx1 << 1, 1, 1;
    vx2 << 1, 0, 1;
    vx3 << 1, 0, 0;

    matrix_type c_diag_1 = matrix_type::Zero(dims, dims);
    c_diag_1.diagonal() = vector_type::Constant(dims, 1);

    matrix_type c_diag_2(dims, dims);
    c_diag_2 << 1.3, 0, 0, 0, 3.4, 0, 0, 0, 23;

    kde_type m0(dims);
    m0.add(explanation_type(v0, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v1, c_diag_1), 1 / 5.0);
    m0.add(explanation_type(v2, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v3, c_diag_2), 1 / 5.0);
    m0.add(explanation_type(v4, c_diag_2), 1 / 5.0);
    m0.debug_set_Nt(5);
    m0.estimate_bandwidth();

    std::vector<int> ev_ok, ev_ko;
    matrix_type F_trns;

    explanation_type smp(m0, false);

    //test smp correctness
    vector_type smp_ref_m(dims);
    smp_ref_m << 3.04, 3.8000000000000007, 38.200000000000003;
    matrix_type smp_ref_cov(dims, dims);
    smp_ref_cov << 26.146400000000007, 5.6879999999999953, 34.871999999999986, 5.6879999999999953, 9.7999999999999936, 83.039999999999992, 34.871999999999986, 83.039999999999992, 3086.7599999999998;
    ASSERT_TRUE(allclose(smp_ref_m, smp.mean()))<< "Ref: " << std::endl
    << smp_ref_m << std::endl << "Actual:" << std::endl << smp.mean();
    ASSERT_TRUE(allclose(smp_ref_cov, smp.covariance()))<< "Ref: " << std::endl << smp_ref_cov << std::endl << "Actual:" << std::endl << smp.covariance();

    smp.get_whitening_parameters(ev_ok, ev_ko, F_trns);

    //test whitening parameters
    matrix_type ref_F_trns(dims, dims);
    ref_F_trns << 0.00020562593514063566, 0.00048529923238563229, 0.017983595023472354, 0.18721070620637167, 0.045842716701635734, -0.0033776790279428401, -0.094172812333361888, 0.38389531308129632, -0.0092828896522654455;
    Eigen::VectorXi ref_ev_ok(dims);
    ref_ev_ok << 0, 1, 2;
    Eigen::VectorXi ref_ev_ko(dims); //real_ev_ko should have dim 0

    //convert the eig_ok and eig_ko into eigen::vector
    int* ptr_ok = &ev_ok[0];
    Eigen::Map<Eigen::VectorXi> real_ev_ok(ptr_ok, ev_ok.size());
    int* ptr_ko = &ev_ko[0];
    Eigen::Map<Eigen::VectorXi> real_ev_ko(ptr_ko, ev_ko.size());

    ASSERT_TRUE(allclose(ref_F_trns,F_trns))<< "Ref: " << std::endl
    << ref_F_trns << std::endl << "Actual:" << std::endl << F_trns;
    ASSERT_TRUE(allclose(ref_ev_ok, real_ev_ok))<< "Ref: " << std::endl
    << ref_ev_ok << std::endl << "Actual:" << std::endl << real_ev_ok;
    ASSERT_TRUE(allclose(ref_ev_ko, real_ev_ko))<< "Ref: " << std::endl
    << ref_ev_ko << std::endl << "Actual:" << std::endl << real_ev_ko;

    kde_type whitened_kde(m0, smp.mean(), ev_ok, ev_ko, F_trns);

    //std::cout << "whitened_kde: " << std::endl << whitened_kde << std::endl;
    vector_type ref_m0(dims), ref_m1(dims);
    ref_m0 << -0.68944256982253693, -0.61429553146616922, -0.81791045549896613;
    ref_m1 << -0.67076804963153824, -0.3846197875861046, -0.53747084440329707;

    matrix_type ref_cov0(dims, dims), ref_cov1(dims, dims);
    ref_cov0 << 0.15781228764380675, -1.4223377250294211e-16, -6.6605249961215751e-16, -1.4369744543579754e-16, 0.19464941206437902, 1.3026689102413336e-16, -6.6586276423197255e-16, 1.0757996056487418e-16, 0.31381890218287001;
    ref_cov1 << 0.15781228764380675, -1.4223377250294211e-16, -6.6605249961215751e-16, -1.4369744543579754e-16, 0.19464941206437902, 1.3026689102413336e-16, -6.6586276423197255e-16, 1.0757996056487418e-16, 0.31381890218287001;

    matrix_type ref_base_cov0(dims, dims), ref_base_cov1(dims, dims);
    ref_base_cov0 << 0.00032368748733841617, 1.3552527156068805e-20, 8.1315162936412833e-20, 1.3552527156068805e-20, 0.037160811908490653, -2.1684043449710089e-18, 8.1315162936412833e-20, -2.1684043449710089e-18, 0.15633030202885725;
    ref_base_cov1 << 0.00032368748733841617, 1.3552527156068805e-20, 8.1315162936412833e-20, 1.3552527156068805e-20, 0.037160811908490653, -2.1684043449710089e-18, 8.1315162936412833e-20, -2.1684043449710089e-18, 0.15633030202885725;

    EXPECT_TRUE(allclose(ref_m0, whitened_kde.component(0).mean())) << "Ref: " << std::endl << ref_m0 << std::endl << "Actual:"
        << std::endl << whitened_kde.component(0).mean();
    EXPECT_TRUE(allclose(ref_m1, whitened_kde.component(1).mean())) << "Ref: " << std::endl << ref_m1 << std::endl << "Actual:"
        << std::endl << whitened_kde.component(1).mean();

    EXPECT_TRUE(allclose(ref_cov0, whitened_kde.component(0).covariance())) << "Ref: " << std::endl << ref_cov0 << std::endl
        << "Actual:" << std::endl << whitened_kde.component(0).covariance();
    EXPECT_TRUE(allclose(ref_base_cov0, whitened_kde.component(0).base_covariance())) << "Ref: " << std::endl << ref_base_cov0
        << std::endl << "Actual:" << std::endl << whitened_kde.component(0).base_covariance();

    EXPECT_TRUE(allclose(ref_cov1, whitened_kde.component(1).covariance())) << "Ref: " << std::endl << ref_cov1 << std::endl
        << "Actual:" << std::endl << whitened_kde.component(1).covariance();
    EXPECT_TRUE(allclose(ref_base_cov1, whitened_kde.component(1).base_covariance())) << "Ref: " << std::endl << ref_base_cov1
        << std::endl << "Actual:" << std::endl << whitened_kde.component(1).base_covariance();

  }

}

