// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#ifndef SRC_INCLUDE_CLASSIFIER_OKDE_MULTI_MODEL_BAYESIAN_H_
#define SRC_INCLUDE_CLASSIFIER_OKDE_MULTI_MODEL_BAYESIAN_H_

#include <map>
#include <string>

#include "kde/basic_types.h"
#include "kde/okde.h"
#include "dataset/split_dataset.h"
#include "dataset/sample.h"
#include "classifier/split_classifier.h"
#include "classifier/prediction.h"

#define DEBUG 1

namespace xokdepp {

  template<typename label_type, typename kde_type>
  class classifier_okde_multi_model_bayesian: public split_classifier<label_type> {

	  typedef sample<label_type> sample_type;

    //each class has its model
    std::map<label_type, kde_type> _models;

    //each class will has its count
    std::map<label_type, double> _counts;

  public:
    classifier_okde_multi_model_bayesian(const split_dataset<label_type> &dataset_to_evaluate) :
        split_classifier<label_type>(dataset_to_evaluate) {

      //    	//initialize _models map
      std::map<label_type, data_type> classes = dataset_to_evaluate.classes();
      for (auto cl : classes)
        _models.insert(std::pair<label_type, kde_type>(cl.first, kde_type(dataset_to_evaluate.dims())));

      train();
      test();
    }

//    std::pair<double, double> models_statistics() const {
//      std::vector<double> model_sizes;
//      for (auto &model : _models) {
//        auto &okde = model.second;
//        model_sizes.push_back(okde.size());
//      }
//
//      double sum_comps = std::accumulate(model_sizes.begin(), model_sizes.end(), 0);
//      double mean = sum_comps / model_sizes.size();
//      double variance = std::accumulate(model_sizes.begin(), model_sizes.end(), 0,
//                                        [&](double acc, double result) {return acc + (result - mean) * (result - mean);});
//      double std_dev = sqrt(variance / model_sizes.size());
//      return std::pair<double, double>(mean, std_dev);
//    }

    std::vector<double> model_complexities() const {
    	std::vector<double> model_sizes;
    	for (auto &model : _models) {
    		auto &okde = model.second;
    		model_sizes.push_back(okde.size());
    	}
    	return model_sizes;
    }

  public:

    friend std::ostream &operator<<(std::ostream &o, const classifier_okde_multi_model_bayesian &c) {
      o << "==========================" << std::endl;
      for (auto mit : c._models) {
        o << "--------------------------" << std::endl;
        o << mit.second << std::endl;
        o << "--------------------------" << std::endl;
      }
      o << "==========================" << std::endl;
      return o;
    }

  private:

    void train(){
    	sequential_train();
    	//parallel_train();
    }

    void sequential_train() {
      auto &train_dataset = this->_dataset_to_evaluate.get_train_dataset();
      for (auto &sample : train_dataset.samples()) {
    	  //std::cout << "Classifier::sample_added" << std::endl;
        _models.at(sample.label()).add_sample(sample.features());
        _counts[sample.label()]++;
      }
      for (auto &model : _models) {
        auto &okde = model.second;
        okde.estimate_kernel_density();
      }
    }

    void parallel_train() {
          auto &train_dataset = this->_dataset_to_evaluate.get_train_dataset();
          size_t num_classes = this->number_of_classes();
          //std::map<label_type, sample_type>  binned_samples = bin_samples(train_dataset);

          //bin samples
          std::vector<std::vector<sample_type>>  binned_samples(num_classes);
          for(auto &sample : train_dataset.samples()){
        	  binned_samples[std::get<1>(sample.label())].push_back(sample);
//        	  std::cout << std::get<1>(sample.label()) << std::endl;
          }
          std::cout << "Binned samples" << std::endl;

#pragma omp parallel for
          for(int i = 0 ; i < num_classes ; i++){
        	  for (auto &sample : binned_samples[i]) {
        		  std::cout << "Classifier::sample_added in thread: " << i << std::endl;

        		  //vector_type truncated_features = sample.features();
        		  //truncated_features.conservativeResize(truncated_features.rows()/TRUCATE_FACTOR);

        		  //vector_type truncated_features = block_crop_img32x32(sample.features(), TRUCATE_FACTOR);
        		  //_models.at(sample.label()).add_sample(truncated_features);

//        		  std::cout << "features" << std::endl << truncated_features << std::endl;
//        		  throw std::exception();
        		  _models.at(sample.label()).add_sample(sample.features());
        		  _counts[sample.label()]++;
        	  }
          }

          for (auto &model : _models) {
            auto &okde = model.second;
            okde.estimate_kernel_density();
          }
        }

    vector_type block_crop_img32x32(const vector_type &features, const size_t truncate_factor){
    	vector_type truncated_features(features.rows()/truncate_factor);
    	size_t head = 0;
    	for(int i=0; i < 8 ; i++){
    		for(int j=0; j < 16 ; j++){
    			truncated_features[head++] = features[i*32 + j];
    		}
    	}
//    	std::cout << "truncated_feats" << std::endl << truncated_features << std::endl;
//    	throw std::exception();
    	return truncated_features;
    }

    void test() {
      const auto &test_dataset = this->_dataset_to_evaluate.get_test_dataset();
      for (auto &sample : test_dataset.samples()) {
        data_type highest_likelihood = 0;
        label_type highest_likelihood_label;

        for (auto &model : _models) {
          auto &okde = model.second;

//          vector_type truncated_features = sample.features();
//          truncated_features.conservativeResize(truncated_features.rows()/8);

          //vector_type truncated_features = block_crop_img32x32(sample.features(), TRUCATE_FACTOR);
          //data_type likelihood = okde.likelihood(truncated_features);

          data_type likelihood = okde.likelihood(sample.features());
          if (likelihood > highest_likelihood) {
            highest_likelihood = likelihood;
            highest_likelihood_label = model.first;
          }
        }

        if (highest_likelihood > 0) {
          this->add_prediction(sample.label(), highest_likelihood_label, highest_likelihood);
        } else {
          this->add_prediction(sample.label());
        }
      }
    }

    //TODO
    void test_using_relative_sample_distribution(){
    	const auto &test_dataset = this->_dataset_to_evaluate.get_test_dataset();
    	for (auto &sample : test_dataset.samples()) {
    		data_type highest_likelihood = 0;
    		label_type highest_likelihood_label;

    		for (auto &model : _models) {
    			auto &okde = model.second;
    			data_type likelihood = okde.likelihood(sample.features());
    			if (likelihood > highest_likelihood) {
    				highest_likelihood = likelihood;
    				highest_likelihood_label = model.first;
    			}
    		}

    		if (highest_likelihood > 0) {
    			this->add_prediction(sample.label(), highest_likelihood_label, highest_likelihood);
    		} else {
    			this->add_prediction(sample.label());
    		}
    	}
    }

  public:
    //computes the sum of the individual likelihoods from all predictions in the model. aka the maximum relative likelihood for the test data
    data_type compute_sum_log_likelihood(){
    	data_type ll_sum = 0;
    	const auto &test_dataset = this->_dataset_to_evaluate.get_test_dataset();

    	for (auto &sample : test_dataset.samples()) {
    		//    	        data_type highest_likelihood = 0;
    		//    	        label_type highest_likelihood_label;

    		for (auto &model : _models) {
    			auto &okde = model.second;
    			//auto &okde =  _models.find(sample.label());
    			if(std::get<0>(model.first) == std::get<0>(sample.label())){
    				double likelihood = okde.likelihood(sample.features());

    				if(likelihood > 1){
    					//std::cout << "Likelihood: " << likelihood << std::endl;
    					std::cerr << "WARNING :: Likelihood: " << std::endl << std::setprecision(9) << okde.likelihood_debug(sample.features()) << std::endl;
    					//std::cout << okde << std::endl;
    					//throw std::exception();
    					continue ; //ignore this sample
    				}

    				if(likelihood > 0.0){
    					ll_sum += log(likelihood);
    				}
    				else{
    					ll_sum += -460.517018599;
    				}
    			}
    		}
    		//}
    	}


//    	for(auto pred : this->predictions()){
//    		//std::cout << "Pred likelihood: " << pred.likelihood() << std::endl;
//    		//std::cout << "log_pred likelihood: " << log(pred.likelihood()) << std::endl;
//    		_models.find(pred.groundtruth_label()).
//    		if(pred.likelihood() > 0.0){
//    			ll_sum += log(pred.likelihood());
//    		}
//    		else{
//    			ll_sum += -460.517018599;
//    		}
//    	}
    	return ll_sum;
    }

    data_type compute_average_negative_log_likelihood(){
    	return -compute_sum_log_likelihood()  / this->number_of_evaluated_samples();
    }

    data_type compute_BIC(){
    	size_t dims = _models.begin()->second.dims();
    	data_type total_num_comps = 0;
    	for (auto &model : _models) {
    		total_num_comps += model.second.size();
    	}

    	data_type f = total_num_comps-1+ total_num_comps*dims + total_num_comps*(dims*(dims-1)/2 + dims) ;
    	return -2*compute_sum_log_likelihood() + f*log(this->number_of_evaluated_samples());
    }

    data_type compute_AIC(){
    	size_t dims = _models.begin()->second.dims();
    	data_type total_num_comps = 0;
    	for (auto &model : _models) {
    		total_num_comps += model.second.size();
    	}

    	data_type f = total_num_comps-1+ total_num_comps*dims + total_num_comps*(dims*(dims-1)/2 + dims) ;
    	data_type AIC =  2*compute_sum_log_likelihood() + 2*f;
    	data_type small_sample_number_correction = ( 2*f*(f+1)) / (this->number_of_evaluated_samples() - f - 1);//need to correct as in high dims we need a lot more samples

    	return AIC + small_sample_number_correction;
    }


  private:

    std::map<label_type, sample_type>   bin_samples(dataset<label_type> &data){
    	size_t num_classes = this->number_of_classes();
    	std::map<label_type, sample_type>  binned_samples;
    	for(auto &sample : data.samples()){
    		binned_samples[sample.label()].insert(sample);
    	}
    	return binned_samples;
    }

  };

}

#endif /* SRC_INCLUDE_CLASSIFIER_OKDE_MULTI_MODEL_BAYESIAN_H_ */
