// (C) Copyright 2014-2016, Jaime Ferreira, David Martins de Matos, Ricardo Ribeiro
// Spoken Language Systems Lab, INESC ID, IST/Universidade de Lisboa

// This file is part of XOKDE++.

// XOKDE++ is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// XOKDE++ is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include <iostream>
#include "kde/okde.h"
#include "kde/gaussian.h"
#include "dataset/load_dsv.h"
#include "dataset/sample.h"
#include "dataset/split_dataset.h"
#include "classifier/classifier_okde_multi_model_bayesian.h"

namespace xokdepp {

#define DATASET xokdepp::iris_data()
  constexpr int SPLIT_RATIO = 75;

  typedef std::tuple<std::string, data_type> label_type;
  typedef split_dataset<label_type> dataset_type;

  typedef oKDE<gaussian<matrix_type>, matrix_type> kde_gaussian_full_type;
  typedef oKDE<gaussian<vector_type>, vector_type> kde_gaussian_diagonal_type;

  //default for tests
  typedef kde_gaussian_diagonal_type kde_type;

  inline std::ostream &operator<<(std::ostream &o, const label_type &lt) {
    o << std::get<0>(lt) << "," << std::get<1>(lt);
    return o;
  }

  inline bool operator==(const label_type &lt1, const label_type &lt2) {
    return std::get<0>(lt1) == std::get<0>(lt2) && std::get<1>(lt1) == std::get<1>(lt2);
  }

  dataset_type iris_data(const char *path = "./data/iris/iris.data", char delim = ',', double ratio = SPLIT_RATIO,
                         size_t nfeats = 4) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type redwine_data(const char *path = "./data/winequality/winequality-red.data", char delim = ';', double ratio =
                                SPLIT_RATIO,
                            size_t nfeats = 11) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type whitewine_data(const char *path = "./data/winequality/winequality-white.data", char delim = ';', double ratio =
                                  SPLIT_RATIO,
                              size_t nfeats = 11) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type yeast_data(const char *path = "./data/yeast/yeast.data", char delim = ',', double ratio = SPLIT_RATIO,
                          size_t nfeats = 8) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type cancer_data(const char *path = "./data/breast_cancer/wdbc.csv", char delim = ',', double ratio = SPLIT_RATIO,
                           size_t nfeats = 30) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type pima_data(const char *path = "./data/pima/pima-indians-diabetes.csv", char delim = ',', double ratio = SPLIT_RATIO,
                         size_t nfeats = 8) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type letter_data(const char *path = "./data/letter/letter_recognition.csv", char delim = ',', double ratio = SPLIT_RATIO,
                           size_t nfeats = 16) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type segmentation_data(const char *path = "./data/segmentation/image_segmentation.csv", char delim = ',', double ratio =
                                     SPLIT_RATIO,
                                 size_t nfeats = 19) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type steel_data(const char *path = "./data/steel/steel_plates.csv", char delim = ',', double ratio = SPLIT_RATIO,
                          size_t nfeats = 27) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type wine_data(const char *path = "./data/wine/wine.csv", char delim = ',', double ratio = SPLIT_RATIO,
                         size_t nfeats = 13) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type faces_pixels_crop16_subset_data(const char *path =
                                                   "./data/faces/pixels/crop16/train_2000_labeled_shuffled_subset50.data",
                                               char delim = ',', double ratio = SPLIT_RATIO, size_t nfeats = 1024) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type faces_pixels_crop16_data(const char *path = "./data/faces/pixels/crop16/train_test_2000_1000_labeled.csv",
                                        char delim = ',', double ratio = 66.66666666, size_t nfeats = 1024) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type faces_pixels_crop0_data(const char *path = "./data/faces/pixels/crop0/train_test_2000_1000_labeled.csv", char delim =
                                           ',',
                                       double ratio = 66.66666666, size_t nfeats = 1024) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type faces_pixels_crop16_128dims_data(const char *path =
                                                    "./data/faces/pixels/crop16/train_test_2000_1000_128dims_labeled.csv",
                                                char delim = ',', double ratio = 66.66666666, size_t nfeats = 128) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  dataset_type custom_data(const char *path, char delim = ',', double ratio = SPLIT_RATIO, size_t nfeats = 0) {
    return load_dsv<label_type, dataset_type>(path, delim, dataset_type(ratio), nfeats);
  }

  template<typename KDE_TYPE = xokdepp::kde_type>
  void run_test_on(dataset_type &&a_dataset) {
    const int NUM_FOLDS = 12;
    std::vector<dataset_type> multiple_dataset_shuffles;

    for (int i = 0; i < NUM_FOLDS; i++) {
      a_dataset.shuffle();
      multiple_dataset_shuffles.push_back(a_dataset);
    }

    std::vector<double> total_correct(NUM_FOLDS), total_incorrect(NUM_FOLDS), total_evaluated(NUM_FOLDS);
    std::vector<double> mean_results(NUM_FOLDS);
    std::vector<std::vector<double>> number_components(NUM_FOLDS);
    double best_result = 0;
    double worst_result = 100;

#pragma omp parallel for
    for (int i = 0; i < NUM_FOLDS; i++) {
      classifier_okde_multi_model_bayesian<label_type, KDE_TYPE> a_classifier(multiple_dataset_shuffles[i]);

      a_classifier.evaluate();
      number_components[i] = a_classifier.model_complexities();
      total_correct[i] = a_classifier.correctly_classified();
      total_incorrect[i] = a_classifier.incorrectly_classified();

      if (total_correct[i] < 10)
        std::cout << a_classifier << std::endl;

      total_evaluated[i] = a_classifier.number_of_evaluated_samples();
      double result = a_classifier.correctly_classified() / a_classifier.number_of_evaluated_samples() * 100;
      mean_results[i] = result;

#pragma omp critical
      {
        if (result > best_result)
          best_result = result;
        if (result < worst_result)
          worst_result = result;
      }
    }

    std::cout << std::endl;
    std::cout << "------------------- Summary -----------------------" << std::endl;
    for (size_t i = 0; i < NUM_FOLDS; i++) {
      std::cout << "--" << std::endl;
      std::cout << "Number of Components: " << std::endl;
      for (double sz : number_components[i]) {
        std::cout << sz << std::endl;
      }
      std::cout << "Total samples: " << total_correct[i] + total_incorrect[i] << std::endl;
      std::cout << "Correctly classified: " << total_correct[i] << " (" << total_correct[i] / total_evaluated[i] * 100 << " %)"
          << std::endl;
      std::cout << "Incorrectly classified: " << total_incorrect[i] << " (" << total_incorrect[i] / total_evaluated[i] * 100
          << " %)" << std::endl;
    }

    std::cout << "------------------- Aggregate Sums -----------------------" << std::endl;
    double sum_correct = std::accumulate(total_correct.begin(), total_correct.end(), 0);
    double sum_incorrect = std::accumulate(total_incorrect.begin(), total_incorrect.end(), 0);
    double sum_evaluated = std::accumulate(total_evaluated.begin(), total_evaluated.end(), 0);
    std::cout << "Total evaluated: " << sum_evaluated << std::endl;
    std::cout << "Total Correctly classified: " << sum_correct << " (" << sum_correct / sum_evaluated * 100 << " %)" << std::endl;
    std::cout << "Incorrectly classified: " << sum_incorrect << " (" << sum_incorrect / sum_evaluated * 100 << " %)" << std::endl;
    std::cout << "------------------- Means -----------------------" << std::endl;
    double sum_means = std::accumulate(mean_results.begin(), mean_results.end(), 0);
    double mean = sum_means / mean_results.size();
    double variance = std::accumulate(mean_results.begin(), mean_results.end(), 0,
                                      [&](double acc, double result) {return acc + (result - mean) * (result - mean);});
    double std_dev = sqrt(variance / mean_results.size());
    std::cout << "Mean classification rate: " << mean << " (+- " << std_dev << ")" << std::endl;
    std::cout << "Best: " << best_result << "        Worst: " << worst_result << std::endl;
    //	std::cout << "Number of components: " << std::endl;

  }

  //matrix_type or vector_type
  template<typename KDE_TYPE>
  void run_simple_test_on(dataset_type &&a_dataset) {
    classifier_okde_multi_model_bayesian<label_type, KDE_TYPE> a_classifier(a_dataset);

    a_classifier.evaluate();
    //std::pair<double,double> number_components = a_classifier.models_statistics();
    double total_correct = a_classifier.correctly_classified();
    double total_incorrect = a_classifier.incorrectly_classified();
    double total_evaluated = a_classifier.number_of_evaluated_samples();

    //TODO this printing should be a function somewhere
    std::cout << std::endl;
    std::cout << "------------------- Summary -----------------------" << std::endl;
    std::cout << "Number of Components: " << std::endl;
    for (double sz : a_classifier.model_complexities()) {
      std::cout << sz << std::endl;
    }
    std::cout << "Total samples: " << total_correct + total_incorrect << std::endl;
    std::cout << "Correctly classified: " << total_correct << " (" << total_correct / total_evaluated * 100 << " %)" << std::endl;
    std::cout << "Incorrectly classified: " << total_incorrect << " (" << total_incorrect / total_evaluated * 100 << " %)"
        << std::endl;
    std::cout << "Average Negative Log Likelihood: " << a_classifier.compute_average_negative_log_likelihood() << std::endl;
    std::cout << "Bayesian Information Criterion (BIC): " << a_classifier.compute_BIC() << std::endl;
    //std::cout << "Aikaike Information Criterion (AIC): " << a_classifier.compute_BIC() << std::endl;
  }

}

int main(int argc, char* argv[]) {
  if (argc == 1) {
    xokdepp::run_test_on(DATASET);
  }
  else if (argc > 1) {
    if(std::string(argv[1]) == "help") {
      std::cout << "Instituto Superior Técnico - Universidade de Lisboa" << std::endl;
      std::cout << "INESC-ID - Laboratório de Língua Falada (L2F)" << std::endl;
      std::cout << "Authors:" << std::endl;
      std::cout << "Jaime Ferreira" << std::endl;
      std::cout << "David Martins de Matos" << std::endl;
      std::cout << "Ricardo Ribeiro" << std::endl;
      std::cout << std::endl << "Usage: " << std::endl;
      std::cout << "If no argument is provided, a default dataset from UCI will be used (iris)" << std::endl;
      std::cout << "Currently available datasets are: " << std::endl;
      std::cout << "iris - (4 dims) - " << std::endl;
      std::cout << "redwine - (11 dims) - " << std::endl;
      std::cout << "whitewine - (11 dims) -  " << std::endl;
      std::cout << "yeast - (8 dims) - " << std::endl;
      std::cout << "cancer - (30 dims) - " << std::endl;
      std::cout << "letter - (16 dims) -  " << std::endl;
      std::cout << "pima - (8 dims) -  " << std::endl;
      std::cout << "segmentation - (19 dims) -  " << std::endl;
      std::cout << "steel - (27 dims) -  " << std::endl;
      std::cout << "wine - (13 dims) -  " << std::endl;
      std::cout << "faces_pixels_crop16_subset - L2f Face Database" << std::endl;
      std::cout << "\t Just 50 samples to show training working (1024 dims) - (pixel features ; image crop of 16px each side)" << std::endl;
      std::cout << "faces_pixels_crop16 - L2f Face Database" << std::endl;
      std::cout << "\t Composed of 30000 samples (1024 dims) - (pixel features ; image crop of 16px each side)" << std::endl;
      std::cout << "faces_pixels_crop16_128dims - L2f Face Database" << std::endl;
      std::cout << "faces_pixels_crop0 - L2f Face Database" << std::endl;
      std::cout << std::endl << "To select a dataset simply give it's name as argument" << std::endl;
      std::cout << "To use an arbitrary dataset use keyword \"file\" and provide the <filename> ; <number of features> and <covariance type> as arguments." << std::endl;
      std::cout << "\t\t  Available covariance types are: full | diagonal " << std::endl;
      std::cout << "\t\t  Example: file my_custom_dataset_with_4_features.csv 4 full" << std::endl;
      //std::cout << "\t Extra arguments are available, such as train/test partition (given in percentage). If not provided default is 75 25" << std::endl;
      //std::cout << "\t\t  Example: file my_custom_dataset.csv 75 25" << std::endl;
    }
    else if(std::string(argv[1]) == "iris") {
      xokdepp::run_test_on(xokdepp::iris_data());
    }
    else if(std::string(argv[1]) == "redwine") {
      xokdepp::run_test_on(xokdepp::redwine_data());
    }
    else if(std::string(argv[1]) == "whitewine") {
      xokdepp::run_test_on(xokdepp::whitewine_data());
    }
    else if(std::string(argv[1]) == "yeast") {
      xokdepp::run_test_on(xokdepp::yeast_data());
    }
    else if(std::string(argv[1]) == "cancer") {
      xokdepp::run_test_on(xokdepp::cancer_data());
    }
    else if(std::string(argv[1]) == "letter") {
      xokdepp::run_test_on(xokdepp::letter_data());
    }
    else if(std::string(argv[1]) == "pima") {
      xokdepp::run_test_on(xokdepp::pima_data());
    }
    else if(std::string(argv[1]) == "segmentation") {
      xokdepp::run_test_on(xokdepp::segmentation_data());
    }
    else if(std::string(argv[1]) == "steel") {
      xokdepp::run_test_on(xokdepp::steel_data());
    }
    else if(std::string(argv[1]) == "wine") {
      xokdepp::run_test_on(xokdepp::wine_data());
    }
    else if(std::string(argv[1]) == "faces_pixels_crop16") {
      xokdepp::run_simple_test_on<xokdepp::kde_gaussian_diagonal_type>(xokdepp::faces_pixels_crop16_data());
    }
    else if(std::string(argv[1]) == "faces_pixels_crop0") {
      xokdepp::run_simple_test_on<xokdepp::kde_gaussian_diagonal_type>(xokdepp::faces_pixels_crop0_data());
    }
    else if(std::string(argv[1]) == "faces_pixels_crop16_128dims") {
      xokdepp::run_simple_test_on<xokdepp::kde_gaussian_diagonal_type>(xokdepp::faces_pixels_crop16_128dims_data());
    }
    else if(std::string(argv[1]) == "file") {
      std::istringstream iss( argv[3] );
      int num_features;
      if (iss >> num_features)
      {
        std::istringstream iss( argv[5] );
        double split_ratio;
        if (iss >> split_ratio)
        {
          //conversion successfull
          std::cout << "Split ratio = " << split_ratio << std::endl;
        }
        else {
          split_ratio = 75;
          std::cout << "Using default split ratio = " << split_ratio << std::endl;
        }

        // Conversion successful
        if(std::string(argv[4]) == "full"){
          xokdepp::run_simple_test_on<xokdepp::kde_gaussian_full_type>(xokdepp::custom_data(argv[2], ',' , split_ratio, num_features));
        }
        else if(std::string(argv[4]) == "diagonal"){
          xokdepp::run_simple_test_on<xokdepp::kde_gaussian_diagonal_type>(xokdepp::custom_data(argv[2], ',' , split_ratio, num_features));
        }
        else {
          std::cout << "Incorrect format of \"covariance type\" argument. See \"help \" for help." << std::endl;
        }
      }
      else {
        std::cout << "Incorrect format of \"number of features\" argument. See \"help \" for help." << std::endl;
      }


    }
    else {
      std::cout << "Please provide a valid dataset. Run with option \"help\" to display help." << std::endl;
    }
  }
  else {
    std::cerr << "Some strange error with program arguments occured. Stopping now." << std::endl;
    std::cerr << "DEBUG - argc should be >= 1" << std::endl;
  }
  return 0;
}

